import pandas as pd
from carrot.io.common import DataCollection,DataBrick
import glob
import io
import os
import json
import pandas as pd
from time import gmtime, strftime

class LocalDataCollection(DataCollection):
    def __init__(self,file_map=None,chunksize=None,nrows=None,output_folder=None,sep=',',write_mode='w',write_separate=False,**kwargs):
        super().__init__(chunksize=chunksize,nrows=nrows)

        self.__output_folder = output_folder
        self.__separator = sep
        self.__write_mode = write_mode
        self.__write_separate = write_separate

        if file_map is not None:
            self._load_input_files(file_map)

    def get_output_folder(self):
        return self.__output_folder

    def get_global_ids(self):
        if not self.__output_folder:
            return

        files = glob.glob(self.__output_folder+os.path.sep+"person_ids.*"+self.get_outfile_extension())
        return files

    def load_global_ids(self):
        if self.__write_mode == 'w':
            return

        files = self.get_global_ids()
        if not files:
            return

        self.logger.warning(f"Loading existing person ids from...")
        self.logger.warning(f"{files}")
        return pd.concat([pd.read_csv(fname,sep=self.__separator).set_index('TARGET_SUBJECT')['SOURCE_SUBJECT']
                          for fname in files
        ]).to_dict()

    def get_separator(self):
        return self.__separator

    def get_outfile_extension(self):
        """
        Work out what the extension of the output file for the dataframes should be.

        Given the '_outfile_separator' to be used in `df.to_csv`,
        work out the file extension.

        At current, only tab separated and comma separated values (files) are supported

        Returns:
           str: outfile extension name

        """
        if self.__separator == ',':
            return 'csv'
        elif self.__separator == '\t':
            return 'tsv'
        else:
            self.logger.warning(f"Don't know what to do with the extension '{self.__separator}' ")
            self.logger.warning("Defaulting to csv")
            return 'csv'


    def load_meta(self,name='.meta'):
        f_out = self.__output_folder
        fname = f"{f_out}{os.path.sep}{name}.json"
        if not os.path.exists(fname):
            return
        with open(fname,'r') as f:
            data = json.load(f)
            return data

    def load_indexing(self):
        meta = self.load_meta()
        if not meta:
            return

        indexing = {}
        for _,v in meta.items():
            v = v['meta']['total_data_processed']
            for k,n in v.items():
                if k not in indexing:
                    indexing[k] = 0
                indexing[k] += n
        return indexing

    def write_tsv_summary(self,data,name='.summary'):
        mode = self.__write_mode
        f_out = self.__output_folder
        if not os.path.exists(f'{f_out}'):
            self.logger.info(f'making output folder {f_out}')
            os.makedirs(f'{f_out}')

        fname = f"{f_out}{os.path.sep}{name}.tsv"
        #rewrite it
        with open(fname,"w") as f:
            f.write(data)
            f.close()

    def write_meta(self,data,name='.meta'):
        if not isinstance(data,dict):
            raise NotImplementedError(f"{type(data)} must be of type dict")

        data = {hex(id(data)):data}

        mode = self.__write_mode
        f_out = self.__output_folder
        if not os.path.exists(f'{f_out}'):
            self.logger.info(f'making output folder {f_out}')
            os.makedirs(f'{f_out}')

        fname = f"{f_out}{os.path.sep}{name}.json"
        if os.path.exists(fname) and mode == 'a':
            with open(fname,'r') as f:
                existing_data = json.load(f)
                data = {**existing_data,**data}
        #rewrite it
        with open(f"{f_out}{os.path.sep}{name}.json","w") as f:
            json.dump(data,f,indent=6)
            return

    def write(self,name,df,mode='w'):

        f_out = self.__output_folder
        if not os.path.exists(f'{f_out}'):
            self.logger.info(f'making output folder {f_out}')
            os.makedirs(f'{f_out}')

        if mode == None:
            mode = self.__write_mode

        if self.__write_separate:
            time = strftime("%Y-%m-%dT%H%M%S", gmtime())
            if 'name' in df.attrs:
                name = name + '.' + df.attrs['name']
            name = name + "."+ hex(id(df)) + "." + time
            mode = 'w'



        file_extension = self.get_outfile_extension()
        fname = f'{f_out}{os.path.sep}{name}.{file_extension}'
        #force mode to write if the file doesnt exist yet
        if not os.path.exists(fname):
            mode = 'w'

        header=True
        if mode == 'a':
            header = False
        if mode == 'w':
            self.logger.info(f'saving {name} to {fname}')
        else:
            self.logger.info(f'updating {name} in {fname}')

        for col in df.columns:
            if col.endswith("_id"):
                df[col] = df[col].astype(float).astype(pd.Int64Dtype())

        df.set_index(df.columns[0],inplace=True)
        self.logger.debug(df.dtypes)
        df.to_csv(fname,mode=mode,header=header,index=True,sep=self.__separator)

        self.logger.debug(df.dropna(axis=1,how='all'))
        self.logger.info("finished save to file")
        return fname

    def _load_input_files(self,file_map):
        for name,path in file_map.items():
            df = pd.read_csv(path,
                             chunksize=self.chunksize,
                             nrows=self.nrows,
                             dtype=str)
            self[name] = DataBrick(df)

    def load_input_dataframe(self,file_map):
        for name,df in file_map.items():
            self[name] = DataBrick(df)



import json
import pandas as pd
import os
from . import templates
import glob
from carrot.cdm import classes

def make_class(data,
               name=None,
               f_inputs=None):
               
    
    structural_mapping = data['cdm']

    metadata = data['metadata']
    person_ids = None
    if 'person_id' in metadata:
        person_ids = metadata['person_id']
        person_ids = { k.lower():v.lower() for k,v in person_ids.items() }

    
    objects = []
    for destination_table,_map in structural_mapping.items():
        for i,(_,obj) in enumerate(_map.items()):
            map_rules = []
            #find structural mapping

            for destination_field,source in sorted(obj.items()):
                source_field = source['source_field']
                source_table = source['source_table']
                rule = templates.rule.render(destination_field=destination_field,
                                             source_table=source_table,
                                             source_field=source_field)
                map_rules.append(rule)


            #add a line break
            map_rules.append('')
            map_rules.append('# --- insert field operations --- ')

            for destination_field,source in sorted(obj.items()):
                if 'operations' in source and source['operations'] is not None:
                    for operation in source['operations']:
                        rule = templates.operation.render(destination_field=destination_field,
                                                          operation=operation)
                        map_rules.append(rule)
                        
                
            #add a line break
            map_rules.append('')
            map_rules.append('# --- insert term mapping --- ')
            #find term mapping
            for destination_field,source in sorted(obj.items()):
                term_mapping = None
                if 'term_mapping' in source:
                    term_mapping = source['term_mapping']

                
                if term_mapping:
                    #term map each value
                    if isinstance(term_mapping,dict):
                        nindent=4
                        term_mapping = json.dumps(term_mapping,indent=nindent).splitlines()
                        temp = f'self.{destination_field}.series = self.{destination_field}.series.map('
                        map_rules.append(temp)
                        for line in term_mapping:
                            map_rules.append(f'{" "*nindent}{line}')
                        map_rules.append(f')')
                    #force all values to be a single value
                    else:
                        temp = f'self.{destination_field}.series = self.tools.make_scalar(self.{destination_field}.series,{term_mapping})'
                        map_rules.append(temp)
                        


            
            function_name = f"{destination_table}_{i}"
            objects.append(templates.obj.render(
                function_name=function_name,
                object_name=destination_table,
                map_rules=map_rules))


    init = templates.init.render(person_ids=person_ids)
    source_code = templates.cls.render(name=name, init=init, objects=objects)

    current_dir = os.getcwd()

    fname = os.path.join(current_dir, f"{name}.py")
    if os.path.isfile(fname):
        print (f"Recreating file {fname}")
    else:
        print (f"Making new file {fname}")
        
    with open(fname,'w') as f:
        f.write(source_code)

    return fname
    
def register_class(fname):
    name = os.path.basename(fname)
    #register the file within carrot/classes so it can be imported
    save_dir = os.path.dirname(os.path.abspath(classes.__file__))

    config_folder = os.environ.get('CONFIG_CONFIG_FOLDER')
    if config_folder is not None:
        save_dir = config_folder
        fname_dst = os.path.join(save_dir, name)
        if fname == fname_dst:
            print (f"File is already present in {config_folder}")
            return
        
    fname_dst = os.path.join(save_dir, name)
    if os.path.isfile(fname_dst):
        os.unlink(fname_dst)
    print (f"Registering file, creating a symlink to {fname_dst}")
    os.symlink(fname, fname_dst)




import pandas as pd
from ...common import DestinationTable, DestinationField

class Specimen(DestinationTable):
    """
    CDM Specimen object class
    """
    
    name = 'specimen'
    def __init__(self,name=None):

        self.specimen_id                 = DestinationField(dtype="Integer"   , required=False , pk=True)
        self.person_id                   = DestinationField(dtype="Integer"   , required=False )
        self.specimen_concept_id         = DestinationField(dtype="Integer"   , required=False )
        self.specimen_type_concept_id    = DestinationField(dtype="Integer"   , required=False )
        self.specimen_date               = DestinationField(dtype="Date"      , required=False )
        self.specimen_datetime           = DestinationField(dtype="Timestamp" , required=False )
        self.quantity                    = DestinationField(dtype="Float"     , required=False )
        self.unit_concept_id             = DestinationField(dtype="Integer"   , required=False )
        self.anatomic_site_concept_id    = DestinationField(dtype="Integer"   , required=False )
        self.disease_status_concept_id   = DestinationField(dtype="Integer"   , required=False )
        self.specimen_source_id          = DestinationField(dtype="Text50"    , required=False )
        self.specimen_source_value       = DestinationField(dtype="Text50"    , required=False )
        self.unit_source_value           = DestinationField(dtype="Text50"    , required=False )
        self.anatomic_site_source_value  = DestinationField(dtype="Text50"    , required=False )
        self.disease_status_source_value = DestinationField(dtype="Text50"    , required=False )
        
        if name is None:
            name = hex(id(self))
        super().__init__(name,self.name)

    def get_df(self,**kwargs):
        """
        Overload/append the creation of the dataframe, specifically for the specimen objects
        * specimen_concept_id is required to be not null
          this can happen when spawning multiple rows from a person
          we just want to keep the ones that have actually been filled
        
        Returns:
           pandas.Dataframe: output dataframe
        """

        df = super().get_df(**kwargs)

        if self.automatically_fill_missing_columns == True:
            if df['specimen_date'].isnull().all():
                df['specimen_date'] = self.tools.get_date(df['specimen_datetime'])
        return df



import os
import click
import json
import carrot
import pandas as pd
import numpy as np
import requests
from dotenv import dotenv_values
    
class MissingToken(Exception):
    pass


@click.group(help='Commands for search/help with mapping')
def search():
    pass



@click.group(help='Commands to get data from the CCOM OMOP database.')
@click.pass_context
def omop(ctx):
    from sqlalchemy import create_engine, inspect
    from sqlalchemy.schema import CreateSchema
    from sqlalchemy_utils import database_exists, create_database

    config = dotenv_values(".env")
    
    host = config['carrot_DB_HOST']
    port = config['carrot_DB_PORT']
    dbname = config['carrot_DB_NAME']
    user = config['carrot_DB_USER']
    password = config['carrot_DB_PASSWORD']
    
    connection_str = f'postgresql://{user}:{password}@{host}/{dbname}'
    
    psql_engine = create_engine(connection_str)

    #insp  = inspect(psql_engine)
    #existing_tables = insp.get_table_names()

    ctx.obj = psql_engine
    

@click.command(help='omop code')
@click.option("--concept-id",help='omop code')
@click.option("--concept-name",help='omop code')
@click.option("--standard",is_flag=True)
@click.pass_obj
def concept(psql_engine,concept_id,concept_name,standard):

    if concept_id:
        query = f"SELECT * FROM concept WHERE concept_id='{concept_id}'"
    elif concept_name:
        #query = f"SELECT * FROM concept WHERE concept_name LIKE '%{concept_name}%'"
        query = f"SELECT * FROM concept WHERE concept_name ILIKE '%%{concept_name}%%'"
    else:
        click.echo("Error... you must used --concept-id or --concept-name")
        return

    if standard:
        query += " and standard_concept='S'"
        
    df = pd.read_sql(query,psql_engine).astype(str)#.squeeze(axis=0)
    #print (df)#json.dumps(df.to_dict(),indent=6))
    #for series in df:
    #    print (series)
    data = df.to_dict(orient='records')
    print (json.dumps(data,indent=6))



omop.add_command(concept,"concept")
search.add_command(omop,"omop")
#search.add_command(mongo,"mongo")





import pandas as pd
from ...common import DestinationTable, DestinationField

class VisitOccurrence(DestinationTable):
    """
    CDM Visit Occurrence object class
    """
    
    name = 'visit_occurrence'
    def __init__(self,name=None):
        self.visit_occurrence_id           = DestinationField(dtype="Integer"   , required=True, pk=True)
        self.person_id                     = DestinationField(dtype="Integer"   , required=True)
        self.visit_concept_id              = DestinationField(dtype="Integer"   , required=True)
        self.visit_start_date              = DestinationField(dtype="Date"      , required=True)
        self.visit_start_datetime          = DestinationField(dtype="Timestamp" , required=False)
        self.visit_end_date                = DestinationField(dtype="Date"      , required=True)
        self.visit_end_datetime            = DestinationField(dtype="Timestamp" , required=False)
        self.visit_type_concept_id         = DestinationField(dtype="Integer"   , required=True)
        self.provider_id                   = DestinationField(dtype="Integer"   , required=False)
        self.care_site_id                  = DestinationField(dtype="Integer"   , required=False)
        self.visit_source_value            = DestinationField(dtype="Text50"    , required=False)
        self.visit_source_concept_id       = DestinationField(dtype="Integer"   , required=False)
        self.admitting_source_concept_id   = DestinationField(dtype="Integer"   , required=False)
        self.admitting_source_value        = DestinationField(dtype="Text50"    , required=False)
        self.discharge_to_concept_id       = DestinationField(dtype="Integer"   , required=False)
        self.discharge_to_source_value     = DestinationField(dtype="Text50"    , required=False)
        self.preceding_visit_occurrence_id = DestinationField(dtype="Integer"   , required=False)
        
        if name is None:
            name = hex(id(self))
        super().__init__(name,self.name)



import click
import requests
import carrot
import hashlib
import os
from carrot.tools.logger import Logger

@click.command(help="Command to help pseudonymise data.")
@click.option("-s","--salt",help="salt hash",required=True,type=str)
@click.option("--person-id","-i","--id",help="name of the person_id",required=True,type=str)
@click.option("--output-folder","-o",help="path of the output folder",required=True,type=str)
@click.option("--chunksize",help="set the chunksize when loading data",type=int,default=None)
@click.argument("input",required=True)
def pseudonymise(input,output_folder,chunksize,salt,person_id):

    logger = Logger("pseudonymise")
    logger.info(f"Working on file {input}, pseudonymising column '{person_id}' with salt '{salt}'")
    
    #create the dir
    os.makedirs(output_folder,exist_ok=True)
    f_out = f"{output_folder}{os.path.sep}{os.path.basename(input)}"

    logger.info(f"Saving new file to {f_out}")
    
    #load data
    data = carrot.tools.load_csv(input,chunksize=chunksize)

    i = 0 
    while True:
        data[input][person_id] =  data[input][person_id].apply(
            lambda x: hashlib.sha256((x+salt).encode("UTF-8")).hexdigest()
        )
        logger.info(data[input][person_id])

        mode = 'w'
        header=True
        if i > 0 :
            mode = 'a'
            header=False
        
        data[input].to_csv(f_out,mode=mode,header=header,index=False)
        log.info(f"Finished {input} of size={len(data[input])} on iteration {i}")
        i+=1
        
        try:
            data.next()
        except StopIteration:
            break
    

    logger.info("Done!")
    return f_out



import pandas as pd
import csv
import inspect
import os, time
import datetime
import fnmatch
import sys
import click
import json
import yaml
import glob
import copy
import subprocess
import cProfile, pstats
import carrot
import carrot.tools as tools

@click.group(help="Commands for mapping data to the OMOP CommonDataModel (CDM).")
def run():
    pass


@click.command(help="Generate a python class from the OMOP mapping json")
@click.option("--name",
              help="give the name of the dataset, this will be the name of the .py class file created")
@click.option("--register",
              is_flag=True,
              help="also register the python config with the tool")
@click.argument("rules")
@click.pass_context

def make_class(ctx,name,rules,register):
    data = tools.load_json(rules)
    if name == None:
        name = data['metadata']['dataset']

    fname = tools.extract.make_class(data,name)
    if register:
        ctx.invoke(register_class,pyconfig=fname)
    return fname

@click.command(help="Register a python class with the tool")
@click.argument("pyconfig")
def register_class(pyconfig):
    pyconfig = os.path.abspath(pyconfig)
    _,conf_extension = os.path.splitext(pyconfig)
    if conf_extension != '.py':
        raise NotImplementedError(f"You're trying to register the file {pyconfig} which isnt a .py file")
    tools.extract.register_class(pyconfig)



@click.command(help="List all the python classes there are available to run")
def list_classes():
    print (json.dumps(tools.get_classes(),indent=6))

@click.command(help="remove a registered class")
@click.pass_context
@click.argument("name")
def remove_class(ctx,name):
    classes = tools.get_classes()
    if name not in classes:
        print (f"{name} is not a registered class")
        ctx.invoke(list_classes)
    else:
        _class = classes[name]
        os.unlink(_class['sympath'])

@click.command(help="Execute the running of the test dataset")
@click.pass_context
def test(ctx):
    _dir = os.path.dirname(os.path.abspath(carrot.__file__))
    _dir = f"{_dir}{os.path.sep}data{os.path.sep}test"
    inputs = glob.glob(f"{_dir}{os.path.sep}inputs{os.path.sep}*.csv")
    rules = f"{_dir}{os.path.sep}rules{os.path.sep}rules_14June2021.json"
    output_folder = 'tests'
    ctx.invoke(run,inputs=inputs,rules=rules,output_folder=output_folder)
    for fname in glob.glob(f"{output_folder}{os.path.sep}*.tsv"):
        tools.diff_csv(fname,fname)


@click.command(help="apply operations (transforms) to a dataset")
@click.option("--config","-c",required=True,type=str)
@click.option("--number-of-rows-per-chunk","--nc",default=1e5,type=int)
@click.option("--output-folder","-o",required=True,type=str)
@click.argument("inputs",nargs=-1,required=True)
def transform(inputs,config,number_of_rows_per_chunk,output_folder):

    logger = tools.logger.Logger("transform")

    if not os.path.exists(output_folder):
        os.makedirs(output_folder)

    if os.path.isfile(config):
        config = json.load(open(config))
    else:
        config = json.loads(config)

    logger.info(json.dumps(config,indent=6))

    inputs = {
        os.path.basename(x):x
        for x in inputs
    }
    input_data = tools.load_csv(inputs,chunksize=number_of_rows_per_chunk)

    operation_tools = carrot.cdm.OperationTools()

    header=True
    mode='w'

    i = 0
    while True:
        if i > 0:
            mode = 'a'
            header=False
        i+=1
        for fname,rules in config.items():
            logger.info(f"Working on {fname}")
            df = input_data[fname]
            for colname,operations in rules.items():
                series = df[colname]
                for operation in operations:
                    logger.info(f".. transforming '{colname}' with operation '{operation}'")
                    fn_operation = operation_tools[operation]
                    series = fn_operation(series)
                df[colname] = series
            fout = f"{output_folder}/{fname}"
            df.to_csv(fout,header=header,mode=mode,index=False)

        try:
            input_data.next()
        except StopIteration:
            break

@click.command()
@click.option("--number-of-rows-per-chunk","--nc",default=1e5,type=int)
@click.option("--output-folder","-o",required=True,type=str)
@click.argument("inputs",nargs=-1,required=True)
def format(inputs,number_of_rows_per_chunk,output_folder):
    """
    Format a CDM model by passing all CDM objects and applying formatting.
    """
    types = list(set([
        os.path.splitext(fname)[1]
        for fname in inputs
    ]))
    if len(types) > 1:
        raise Exception(f"Running with mixed input files '{types}'. Only input tsv or csv files.")
    types = types[0]

    inputs = {
        os.path.basename(x):x
        for x in inputs
    }
    if types == '.csv':
        inputs = tools.load_csv(inputs,chunksize=number_of_rows_per_chunk)
    else:
        inputs = tools.load_tsv(inputs,chunksize=number_of_rows_per_chunk)

    cdm = carrot.cdm.CommonDataModel.from_existing(inputs=inputs,
                                                      output_folder=output_folder,
                                                      format_level=1)
    #cdm.save_files = False
    cdm.process()
    cdm.end()

@click.command()
@click.option("--output-folder","-o",required=True,type=str,help='specify an output folder of where to put the merged tables')
@click.argument("inputs",nargs=-1,required=True)
def merge(inputs,output_folder):
    """
    Build a CDM model from subfiles
    """
    types = list(set([
        os.path.splitext(fname)[1]
        for fname in inputs
    ]))
    if len(types) > 1:
        raise Exception(f"Running with mixed input files '{types}'. Only input tsv or csv files.")
    types = types[0]

    inputs = {
        os.path.basename(x):x
        for x in inputs
    }
    if types == '.csv':
        inputs = tools.load_csv(inputs)#,chunksize=number_of_rows_per_chunk)
    else:
        inputs = tools.load_tsv(inputs)#,chunksize=number_of_rows_per_chunk)

    outputs = carrot.tools.create_csv_store(output_folder=output_folder,sep='\t',write_mode='w',write_separate=False)

    cdm = carrot.cdm.CommonDataModel.from_existing(inputs=inputs,
                                                      do_mask_person_id=False,
                                                      drop_duplicates=True,
                                                      format_level=0,
                                                      outputs=outputs)
    cdm.process()


@click.command(help="loading")
@click.option("--config","-c",required=True,type=str)
@click.argument("inputs",nargs=-1,required=True)
def load(inputs,config):
    inputs = {os.path.basename(x).split('.')[0]:x for x in inputs}
    config = json.loads(config)
    if 'bclink' in config:
        outputs = carrot.io.BCLinkDataCollection(config['bclink'])
    else:
        raise NotImplementedError(f"Not setup to implement configuration for load {config}")

    for destination_table,fname in inputs.items():
        outputs.load(fname,destination_table)

    outputs.finalise()

@click.command()
@click.option("--input",
              required=True,
              help='input csv file')
@click.option("--column",
              required=True,
              help="which column of the input data to modify")
@click.option("--operation",
              required=True,
              help="which operation to apply to the column")
def format_input_data(column,operation,input):
    """
    Useful formatting command for applying an operation on some data before being passed into the ETL-Tool

    """
    optools = carrot.cdm.OperationTools()
    allowed_operations = optools.keys()
    if operation not in allowed_operations:
        raise Exception(f"Operation '{operation}' is not a known operation. Choose from {allowed_operations}")
    df_input = pd.read_csv(input)
    df_input[column] = optools[operation](df_input[column])
    n = 5 if len(df_input) > 5 else len(df_input)
    print (df_input[column].sample(n))
    df_input.to_csv(input)


def _condor(name,commands,jobscript):
    #hold_jid {name} - wait for previous job
    def wrapper(args):
        script = jobscript + '\n' + ' '.join(commands)
        fname = f"{name}.sh"
        with open(fname,"w") as f:
            f.write(script)
        output = subprocess.check_output(['qsub','-N',name,fname]).decode()
        return output
    return wrapper

from importlib import import_module

@click.command()
@click.option("analysis_name","--analysis",default=None,type=str)
@click.argument("analysis_file")
@click.pass_context
def analysis(ctx,analysis_file,analysis_name):
    """
    Use this command to run analyses on input data (in the CDM format) given a configuration yaml file
    """
    _dir = os.path.dirname(analysis_file)
    fname = os.path.splitext(os.path.basename(analysis_file))[0]
    sys.path.append(_dir)
    module = import_module(fname)
    clsmembers = inspect.getmembers(module, inspect.isclass)
    name,cls = clsmembers[0]
    obj = cls()
    if analysis_name:
        ana = obj.get_analysis(analysis_name)
        res = obj.run_analysis(ana)
        print (res)
    else:
        #print (obj.get_analyses())
        #exit(0)
        obj.run_analyses()


@click.command()
@click.option("--max-workers",default=None,type=int)
@click.option("--batch",default=None,type=click.Choice(['condor']))
@click.option("analysis_names","--analysis-name",default=None,multiple=True,type=str)
@click.argument("config")
@click.pass_context
def ___analysis(ctx,config,analysis_names,max_workers,batch):
    """
    Use this command to run analyses on input data (in the CDM format) given a configuration yaml file
    """
    from importlib import import_module
    fname = config
    stream = open(config)
    config = yaml.safe_load(stream)

    inputs = config['cdm']
    inputs = carrot.tools.load_tsv(config['cdm'],
                                      dtype=None)

    cdm = carrot.cdm.CommonDataModel.load(inputs=inputs)

    analyses = config['analyses']

    jobscript = config['condor']['jobscript']


    if analysis_names:
        temp = copy.copy(analyses)
        for name in temp:
            if name not in analysis_names:
                analyses.pop(name)

    for name,_def in analyses.items():
        analysis = _def['analysis']
        cohort = _def['filter']
        if batch:
            if batch == 'condor':
                commands = copy.copy(sys.argv)
                commands.remove('--batch')
                commands.remove('condor')
                commands.extend(['--analysis-name',name])

                f = _condor(name=name,commands=commands,jobscript=jobscript)
            else:
                raise NotImplementedError(f"{batch} mode for --batch not a thing")
        else:
            func = import_module(analysis)
            f = func.create_analysis(cohort)
        cdm.add_analysis(f)

    results = cdm.run_analyses(max_workers=max_workers)



@click.command()
@click.option("--rules",
              required=True,
              help="input json file containing all the mapping rules to be applied")
@click.option("--indexing-conf",
              default=None,
              help="configuration file to specify how to start the indexing")
@click.option("--csv-separator",
              default='\t',
              type=click.Choice([';',':','\t',',',' ',]),
              help="choose a separator to use when dumping output csv files")
@click.option("--use-profiler",
              is_flag=True,
              help="turn on saving statistics for profiling CPU and memory usage")
@click.option("format_level","--format-level",
              default='1',
              type=click.Choice(['0','1','2']),
              help="Choose the level of formatting to apply on the output data. 0 - no formatting. 1 - automatic formatting. 2 (default) - check formatting (will crash if input data is not already formatted).")
@click.option("--output-folder",
              default=None,
              help="define the output folder where to dump csv files to")
@click.option("--write-mode",
              default='w',
              type=click.Choice(['w','a']),
              help="force the write-mode on existing files")
@click.option("--split-outputs",
              is_flag=True,
              help="force the output files to be split into separate files")
@click.option("--allow-missing-data",
              is_flag=True,
              help="don't crash if there is data tables in rules file that hasnt been loaded")
@click.option("output_database","--database",
              default=None,
              help="define the output database where to insert data into")
@click.option("-nc","--number-of-rows-per-chunk",
              default=-1,
              help="Choose the number of rows (INTEGER) of input data to load (chunksize). The option 'auto' will work out the ideal chunksize. Inputing a value <=0 will turn off data chunking (default behaviour).")
@click.option("-np","--number-of-rows-to-process",
              default=None,
              type=int,
              help="the total number of rows to process")
@click.option("--person-id-map",
              default=None,
              help="pass the location of a file containing existing masked person_ids")
@click.option("--db",
              default=None,
              help="instead, pass a connection string to a db")
@click.option("--merge-output",
              is_flag=True,
              help="merge the output into one file")
@click.option("no_mask_person_id","--parse-original-person-id",
              is_flag=True,
              help="turn off automatic conversion (creation) of person_id to (as) Integer")
@click.option("dont_automatically_fill_missing_columns","--no-fill-missing-columns",
              is_flag=True,
              help="Turn off automatically filling missing CDM columns")
@click.option("log_file","--log-file",
              default = 'none',
              help="specify a path for a log file")
@click.option("--max-rules",
              default = None,
              type=int,
              help="maximum number of rules to process")
@click.option("objects","--object",
              default = None,
              multiple=True,
              type=str,
              help="give a list of objects by name to process")
@click.option("tables","--table",
              default = None,
              multiple=True,
              type=str,
              help="give a list of tables by name to process")
@click.argument("inputs",
                required=False,
                nargs=-1)
@click.pass_context
def map(ctx,rules,inputs,format_level,
        output_folder,output_database,
        csv_separator,use_profiler,log_file,
        no_mask_person_id,indexing_conf,
        person_id_map,max_rules,merge_output,
        objects,tables,db,write_mode,split_outputs,
        dont_automatically_fill_missing_columns,
        number_of_rows_per_chunk,allow_missing_data,
        number_of_rows_to_process):
    """
    Perform OMOP Mapping given an json file and a series of input files

    INPUTS should be a space separated list of individual input files or directories (which contain .csv files)
    """

    if output_folder is None:
        output_folder = f'{os.getcwd()}{os.path.sep}output_data{os.path.sep}'

    #if log_file == 'auto' and carrot.params['log_file'] is None:
    if log_file == 'auto':
        log_file = f"{output_folder}{os.path.sep}logs{os.path.sep}carrot.log"
        carrot.params['log_file'] = log_file
    elif log_file == 'none':
        pass
    else:
        carrot.params['log_file'] = log_file

    #load the json loads
    if type(rules) == dict:
        config = rules
    else:
        config = tools.load_json(rules)

    if tables:
        tables = list(set(tables))
        config = carrot.tools.filter_rules_by_destination_tables(config,tables)

    if objects:
        objects = list(set(objects))
        config = carrot.tools.filter_rules_by_object_names(config,objects)

    if max_rules:
        i = 0
        n = max_rules
        new = {}
        for destination_table,rule_set in config['cdm'].items():
            if destination_table == 'person':
                new[destination_table] = rule_set
            else:
                for name,_rules in rule_set.items():
                    if i>=n:
                        break
                    if destination_table not in new:
                        new[destination_table] = {}
                    new[destination_table][name] = _rules
                    i+=1

        config['cdm'] = new

    name = config['metadata']['dataset']

    if indexing_conf is not None:
        if isinstance(indexing_conf,dict):
            pass
        elif indexing_conf.endswith(".json") and os.path.exists(indexing_conf):
            indexing_conf = tools.load_json(indexing_conf)
        elif indexing_conf.endswith(".csv") and os.path.exists(indexing_conf):
            try:
                indexing_conf = pd.read_csv(indexing_conf,header=None,index_col=0)[1].to_dict()
            except pd.errors.EmptyDataError:
                indexing_conf = None
                pass


    #automatically calculate the ideal chunksize
    if number_of_rows_per_chunk == 'auto':
        #get the fields that are going to be used/loaded
        used_fields = tools.get_mapped_fields_from_rules(config)
        #calculate the number of fields that are to be used per dataset
        n_used_fields = [ len(sublist) for sublist in used_fields.values() ]
        #find what's the largest number of fields loaded by any dataset
        max_n_used_fields = max(n_used_fields)
        #get the number of files used
        n_files = len(n_used_fields)

        # If there is one dataset and one column being used, the max loaded to memory
        #   is 2million rows (this is fairly arbitrary)
        #   it is an approximation assuming the data in the values is relatively small
        #   this should keep the memory usage down
        # When there is more fields and more files loaded, reduce the of rows per chunk
        max_n_rows = 2e6
        number_of_rows_per_chunk = int(max_n_rows/(max_n_used_fields*n_files))
    else:
        try:
            number_of_rows_per_chunk = int(number_of_rows_per_chunk)
        except ValueError:
            raise ValueError(f"number_of_rows_per_chunk must be an Integer or 'auto', you inputted '{number_of_rows_per_chunk}'")

        #turn off chunking if 0 or negative chunksizes are given
        if number_of_rows_per_chunk <= 0 :
            number_of_rows_per_chunk = None

    #check if exists
    if any('*' in x for x in inputs):
        data_dir = os.path.dirname(carrot.__file__)
        data_dir = f'{data_dir}{os.path.sep}data{os.path.sep}'

        new_inputs = []
        for i,x in enumerate(inputs):
            if not os.path.exists(x):
                new_inputs.extend(glob.glob(f"{data_dir}{os.path.sep}{x}"))
            else:
                new_inputs.append(x)
        inputs = new_inputs

    inputs = list(inputs)

    for x in inputs:
        if os.path.isdir(x):
            inputs.remove(x)
            inputs.extend(glob.glob(f'{x}{os.path.sep}*.csv'))

    #convert the list into a map between the filename and the full path
    inputs = {
        os.path.basename(x):x
        for x in inputs
    }

    if db:
        inputs = tools.load_sql(connection_string=db,chunksize=number_of_rows_per_chunk,nrows=number_of_rows_to_process)
    else:
        if allow_missing_data:
            config = carrot.tools.remove_missing_sources_from_rules(config,inputs)

        inputs = tools.load_csv(inputs,
                                rules=config,
                                chunksize=number_of_rows_per_chunk,
                                nrows=number_of_rows_to_process)

    #do something with
    #person_id_map

    if isinstance(output_database,dict):
        if 'bclink' in output_database:
            outputs = carrot.tools.create_bclink_store(bclink_settings=output_database['bclink'],
                                                          output_folder=output_database['cache'],
                                                          sep=csv_separator,
                                                          write_separate=split_outputs,
                                                          write_mode=write_mode)
        else:
            raise NotImplementedError(f"dont know how to configure outputs... {output_database}")
    elif output_database == None:
        outputs = carrot.tools.create_csv_store(output_folder=output_folder,
                                                   sep=csv_separator,
                                                   write_separate=split_outputs,
                                                   write_mode=write_mode)
    else:
        outputs = carrot.tools.create_sql_store()

    #build an object to store the cdm
    cdm = carrot.cdm.CommonDataModel(name=name,
                                        inputs=inputs,
                                        format_level=format_level,
                                        do_mask_person_id=not no_mask_person_id,
                                        outputs = outputs,
                                        #output_folder=output_folder,
                                        #output_database=output_database,
                                        automatically_fill_missing_columns=not dont_automatically_fill_missing_columns,
                                        use_profiler=use_profiler)
    #allow the csv separator to be changed
    #the default is tab (\t) separation
    #if not csv_separator is None:
    #    cdm.set_csv_separator(csv_separator)
    cdm.create_and_add_objects(config)

    cdm.process(conserve_memory=True)
    cdm.close()

    if merge_output:
        ctx.invoke(merge,
                   inputs=glob.glob(f"{output_folder}{os.path.sep}*"),
                   output_folder=output_folder)
@click.command()
@click.option("--rules-file",
              required=True,
              help="json file containing mapping rules")
@click.option("--output-dir",
              default=None,
              help="define the output directory for OMOP-format tsv files")
@click.option("--write-mode",
              default='w',
              type=click.Choice(['w','a']),
              help="force write-mode on output files")
@click.option("--person-file",
              required=True,
              help="File containing person_ids in the first column")
@click.option("--omop-ddl-file",
              required=True,
              help="File containing OHDSI ddl statements for OMOP tables")
@click.option("--omop-config-file",
              required=True,
              help="File containing additional / override json config for omop outputs")
@click.option("--saved-person-id-file",
              default=None,
              required=False,
              help="Full path to person id file used to save person_id state and share person_ids between data sets")
@click.option("--use-input-person-ids",
              required=False,
              default='N',
              help="Use person ids as input without generating new integers")
@click.option("--last-used-ids-file",
              default=None,
              required=False,
              help="Full path to last used ids file for OMOP tables - format: tablename\tlast_used_id, \nwhere last_used_id must be an integer")
@click.option("--log-file-threshold",
              required=False,
              default=0,
              help="Lower outcount limit for logfile output")
@click.argument("input-dir",
                required=False,
                nargs=-1)
def mapstream(rules_file, output_dir, write_mode, person_file, omop_ddl_file, omop_config_file, saved_person_id_file, use_input_person_ids, last_used_ids_file, log_file_threshold, input_dir):
    """
    Map to output using input streams
    """
    print(input_dir)
    if os.path.isdir(input_dir[0]) == False:
        print("Not a directory, input dir {0}".format(input_dir[0]))
        sys.exit(1)

    if os.path.isdir(output_dir) == False:
        print("Not a directory, output dir {0}".format(output_dir))
        sys.exit(1)

    if saved_person_id_file == None:
        saved_person_id_file = output_dir + "/" + "person_ids.tsv"
        if os.path.exists(saved_person_id_file):
            os.remove(saved_person_id_file)
   
    starttime = time.time()
    omopcdm = tools.omopcdm.OmopCDM(omop_ddl_file, omop_config_file)
    #print(omopcdm.dump_ddl())
    mappingrules = tools.mappingrules.MappingRules(rules_file, omopcdm)
    metrics = tools.metrics.Metrics(mappingrules.get_dataset_name(), log_file_threshold)
    nowtime = time.time()

    print("--------------------------------------------------------------------------------")
    print("Loaded mapping rules from: {0} after {1:.5f} secs".format(rules_file, (nowtime - starttime)))
    output_files = mappingrules.get_all_outfile_names()
    record_numbers = {}
    for output_file in output_files:
        record_numbers[output_file] = 1

    fhd = {}
    tgtcolmaps = {}

    try:
        # Add in a saved-person-file existence test and reload from it is necessary returning the last used integer
        if os.path.isfile(saved_person_id_file):
            person_lookup, last_used_integer = load_saved_person_ids(saved_person_id_file)
        else:
            person_lookup = {}
            last_used_integer = 1
        if last_used_ids_file != None:
            if os.path.isfile(last_used_ids_file):
                record_numbers = load_last_used_ids(last_used_ids_file, record_numbers)
 
        #fhp = open(person_file, mode="r", encoding="utf-8-sig")
        #csvrp = csv.reader(fhp)
        person_lookup, rejected_person_count = load_person_ids(person_file, person_lookup, mappingrules, use_input_person_ids, last_used_integer)
        fhpout = open(saved_person_id_file, mode="w")
        fhpout.write("SOURCE_SUBJECT\tTARGET_SUBJECT\n")
        for person_id, person_assigned_id in person_lookup.items():
            fhpout.write("{0}\t{1}\n".format(str(person_id), str(person_assigned_id)))
        fhpout.close()
        for tgtfile in output_files:
            fhd[tgtfile] = open(output_dir + "/" + tgtfile + ".tsv", mode=write_mode)
            if write_mode == 'w':
                outhdr = omopcdm.get_omop_column_list(tgtfile)
                fhd[tgtfile].write("\t".join(outhdr) + "\n")
            tgtcolmaps[tgtfile] = omopcdm.get_omop_column_map(tgtfile)

    except IOError as e:
        print("I/O - error({0}): {1} -> {2}".format(e.errno, e.strerror, str(e)))
        exit()

    print("person_id stats: total loaded {0}, reject count {1}".format(len(person_lookup), rejected_person_count))

    # TODO get this list of input files from the  parsed rules
    existing_input_files = fnmatch.filter(os.listdir(input_dir[0]), '*.csv')
    rules_input_files = mappingrules.get_all_infile_names()
    for infile in existing_input_files:
        if infile not in rules_input_files:
            msg = "ERROR: no mapping rules found for existing input file - {0}".format(infile)
            print(msg)
            metrics.add_log_data(msg)
    for infile in rules_input_files:
        if infile not in existing_input_files:
            msg = "ERROR: no data for mapped input file - {0}".format(infile)
            print(msg)
            metrics.add_log_data(msg)
    rejidcounts = {}
    rejdatecounts = {}
    #src_tgt_counts = {}
    print(rules_input_files)

    for srcfilename in rules_input_files:
        rejidcounts[srcfilename] = 0
        rejdatecounts[srcfilename] = 0

    for srcfilename in rules_input_files:
        outcounts = {}
        rejcounts = {}
        rcount = 0

        try:
            fh = open(input_dir[0] + "/" + srcfilename, mode="r", encoding="utf-8-sig")
            csvr = csv.reader(fh)
        except IOError as e:
            print("Unable to open: {0}".format(input_dir[0] + "/" + srcfilename))
            print("I/O error({0}): {1}".format(e.errno, e.strerror))
            continue

        tgtfiles, src_to_tgt = mappingrules.parse_rules_src_to_tgt(srcfilename)
        infile_datetime_source, infile_person_id_source = mappingrules.get_infile_date_person_id(srcfilename)
        for tgtfile in tgtfiles:
            outcounts[tgtfile] = 0
            rejcounts[tgtfile] = 0
        datacolsall = []
        hdrdata = next(csvr)
        dflist = mappingrules.get_infile_data_fields(srcfilename)
        for colname in hdrdata:
            datacolsall.append(colname)
        inputcolmap = omopcdm.get_column_map(hdrdata)
        pers_id_col = inputcolmap[infile_person_id_source]
        datetime_col = inputcolmap[infile_datetime_source]
        print("--------------------------------------------------------------------------------")
        print("Processing input: {0}".format(srcfilename))
#        print("Processing input: {0}, All input cols = {1}, Data cols = {2}".format(srcfilename, str(datacolsall), str(dflist)))

        for indata in csvr:
            #indata = inputline.strip().split(",")
            key = srcfilename + "~all~all~all~"
            metrics.increment_key_count(key, "input_count")
            rcount += 1
            strdate = indata[datetime_col].split(" ")[0]
            fulldate = parse_date(strdate)
            if fulldate != None:
                #fulldate = "{0}-{1:02}-{2:02}".format(dt.year, dt.month, dt.day)
                indata[datetime_col] = fulldate
            else:
                metrics.increment_key_count(key, "invalid_date_fields")
                continue

            for tgtfile in tgtfiles:
                tgtcolmap = tgtcolmaps[tgtfile]
                auto_num_col = omopcdm.get_omop_auto_number_field(tgtfile)
                pers_id_col = omopcdm.get_omop_person_id_field(tgtfile)

                datacols = datacolsall
                if tgtfile in dflist:
                    datacols = dflist[tgtfile]

                for datacol in datacols:
                    built_records, outrecords, metrics = get_target_records(tgtfile, tgtcolmap, src_to_tgt, datacol, indata, inputcolmap, srcfilename, omopcdm, metrics)
                    if built_records == True:
                        for outrecord in outrecords:
                            if auto_num_col != None:
                                outrecord[tgtcolmap[auto_num_col]] = str(record_numbers[tgtfile])
                                record_numbers[tgtfile] += 1
                            if (outrecord[tgtcolmap[pers_id_col]]) in person_lookup:
                                outrecord[tgtcolmap[pers_id_col]] = person_lookup[outrecord[tgtcolmap[pers_id_col]]]
                                outcounts[tgtfile] += 1
                                key = srcfilename + "~all~all~all~"
                                metrics.increment_key_count(key, "output_count")
                                key = "all~all~" + tgtfile + "~all~"
                                metrics.increment_key_count(key, "output_count")
                                key = srcfilename + "~all~" + tgtfile + "~all~"
                                metrics.increment_key_count(key, "output_count")
                                if tgtfile == "person":
                                    key = srcfilename + "~all~" + tgtfile + "~" + outrecord[1] +"~"
                                    metrics.increment_key_count(key, "output_count")
                                    key = srcfilename + "~" + datacol +"~" + tgtfile + "~" + outrecord[1] + "~" + outrecord[2]
                                    metrics.increment_key_count(key, "output_count")
                                else:
                                    key = srcfilename + "~" + datacol +"~" + tgtfile + "~" + outrecord[2] + "~"
                                    metrics.increment_key_count(key, "output_count")
                                    key = srcfilename + "~all~" + tgtfile + "~" + outrecord[2] + "~"
                                    metrics.increment_key_count(key, "output_count")
                                    key = "all~all~" + tgtfile + "~" + outrecord[2] + "~"
                                    metrics.increment_key_count(key, "output_count")
                                    key = "all~all~all~" + outrecord[2] + "~"
                                    metrics.increment_key_count(key, "output_count")
                                fhd[tgtfile].write("\t".join(outrecord) + "\n")
                            else:
                                key = srcfilename + "~all~" + tgtfile + "~all~"
                                metrics.increment_key_count(key, "invalid_person_ids")
                                rejidcounts[srcfilename] += 1

        fh.close()

        nowtime= time.time()
        print("INPUT file data : {0}: input count {1}, time since start {2:.5} secs".format(srcfilename, str(rcount), (nowtime - starttime)))
        for outtablename, count in outcounts.items():
            print("TARGET: {0}: output count {1}".format(outtablename, str(count)))

    print("--------------------------------------------------------------------------------")
    data_summary = metrics.get_mapstream_summary()
    log_report = metrics.get_log_data()
    try:
        dsfh = open(output_dir + "/summary_mapstream.tsv", mode="w")
        dsfh.write(data_summary)
        dsfh.close()
        logfh = open(output_dir + "/error_report.txt", mode="w")
        logfh.write(log_report)
        logfh.close()
    except IOError as e:
        print("I/O error({0}): {1}".format(e.errno, e.strerror))
        print("Unable to write file")

    nowtime = time.time()
    print("Elapsed time = {0:.5f} secs".format(nowtime - starttime))
    #profiler.disable()
    #stats = pstats.Stats(profiler).sort_stats('ncalls')
    #stats.print_stats()

def get_target_records(tgtfilename, tgtcolmap, rulesmap, srcfield, srcdata, srccolmap, srcfilename, omopcdm, metrics):
    build_records = False
    tgtrecords = []
    date_col_data = omopcdm.get_omop_datetime_linked_fields(tgtfilename)
    date_component_data = omopcdm.get_omop_date_field_components(tgtfilename)
    notnull_numeric_fields = omopcdm.get_omop_notnull_numeric_fields(tgtfilename)

    srckey = srcfilename + "~" + srcfield + "~" + tgtfilename
    summarykey = srcfilename + "~" + srcfield + "~" + tgtfilename + "~all~"
    if valid_value(str(srcdata[srccolmap[srcfield]])):
        srcfullkey = srcfilename + "~" + srcfield + "~" + str(srcdata[srccolmap[srcfield]]) + "~" + tgtfilename
        dictkeys = []
        if srcfullkey in rulesmap:
            build_records = True
            dictkeys.append(srcfullkey)
        if srckey in rulesmap:
            build_records = True
            dictkeys.append(srckey)
        if build_records == True:
            for dictkey in dictkeys:
                for out_data_elem in rulesmap[dictkey]:
                    valid_data_elem = True
                    tgtarray = ['']*len(tgtcolmap)
                    for req_integer in notnull_numeric_fields:
                        tgtarray[tgtcolmap[req_integer]] = "0"
                    for infield, outfield_list in out_data_elem.items():
                        for output_col_data in outfield_list:
                            if "~" in output_col_data:
                                outcol, term = output_col_data.split("~")
                                tgtarray[tgtcolmap[outcol]] = term
                            else:
                                tgtarray[tgtcolmap[output_col_data]] = srcdata[srccolmap[infield]]
                            if output_col_data in date_component_data:
                                strdate = srcdata[srccolmap[infield]].split(" ")[0]
                                dt = get_datetime_value(strdate)
                                if dt != None:
                                    year_field = date_component_data[output_col_data]["year"]
                                    month_field = date_component_data[output_col_data]["month"]
                                    day_field = date_component_data[output_col_data]["day"]
                                    tgtarray[tgtcolmap[year_field]] = str(dt.year)
                                    tgtarray[tgtcolmap[month_field]] = str(dt.month)
                                    tgtarray[tgtcolmap[day_field]] = str(dt.day)
                                    fulldate = "{0}-{1:02}-{2:02}".format(dt.year, dt.month, dt.day)
                                    tgtarray[tgtcolmap[output_col_data]] = fulldate
                                else:
                                    metrics.increment_key_count(summarykey, "invalid_date_fields")
                                    valid_data_elem = False
                            elif output_col_data in date_col_data:
                                fulldate = srcdata[srccolmap[infield]]
                                tgtarray[tgtcolmap[output_col_data]] = fulldate
                                tgtarray[tgtcolmap[date_col_data[output_col_data]]] = fulldate
                    if valid_data_elem == True:
                        tgtrecords.append(tgtarray)
    else:
        metrics.increment_key_count(summarykey, "invalid_source_fields")


    return build_records, tgtrecords, metrics

def valid_value(item):
    """
    Check if an item is non blank (null)
    """
    if item.strip() == "":
        return(False)
    return(True)

def valid_date_value(item):
    """
    Check if a date item is non null and parses as ISO (YYYY-MM-DD), reverse-ISO
    or dd/mm/yyyy or mm/dd/yyyy
    """
    if item.strip() == "":
        return(False)
    if not valid_iso_date(item) and not valid_reverse_iso_date(item) and not valid_uk_date(item):
        #print("Bad date : {0}".format(item))
        return(False)
    return(True)

def get_datetime_value(item):
    """
    Check if a date item is non null and parses as ISO (YYYY-MM-DD), reverse-ISO
    or dd/mm/yyyy or mm/dd/yyyy
    """
    dt = None
    # Does the date parse as an ISO date?
    try:
        dt = datetime.datetime.strptime(item, "%Y-%m-%d")
    except ValueError:
        pass
    if dt != None:
      return(dt)

    # Does the date parse as a reverse ISO date?
    try:
        dt = datetime.datetime.strptime(item, "%d-%m-%Y")
    except ValueError:
        pass

    if dt != None:
      return(dt)

    # Does the date parse as a UK old-style date?
    try:
        dt = datetime.datetime.strptime(item, "%d/%m/%Y")
    except ValueError:
        pass

    if dt != None:
      return(dt)

    return None

def parse_date(item):
    """
    Crude hand-coded check on date format
    """
    datedata = item.split("-")
    if len(datedata) != 3:
        datedata = item.split("/")
    if len(datedata) != 3:
        return None
    if len(datedata[2]) == 4:
        return("{0}-{1}-{2}".format(datedata[2], datedata[1], datedata[0]))
    return("{0}-{1}-{2}".format(datedata[0], datedata[1], datedata[2]))


def valid_iso_date(item):
    """
    Check if a date item is non null and parses as ISO (YYYY-MM-DD)
    """
    try:
        datetime.datetime.strptime(item, "%Y-%m-%d")
    except ValueError:
        return(False)

    return(True)

def valid_reverse_iso_date(item):
    """
    Check if a date item is non null and parses as reverse ISO (DD-MM-YYYY)
    """
    try:
        datetime.datetime.strptime(item, "%d-%m-%Y")
    except ValueError:
        return(False)

    return(True)

def valid_uk_date(item):
    """
    Check if a date item is non null and parses as UK format (DD/MM/YYYY)
    """
    try:
        datetime.datetime.strptime(item, "%d/%m/%Y")
    except ValueError:
        return(False)

    return(True)

def load_last_used_ids(last_used_ids_file, last_used_ids):
    fh = open(last_used_ids_file, mode="r", encoding="utf-8-sig")
    csvr = csv.reader(fh, delimiter="\t")

    for last_ids_data in csvr:
        last_used_ids[last_ids_data[0]] = int(last_ids_data[1]) + 1

    fh.close()
    return last_used_ids

def load_saved_person_ids(person_file):
    fh = open(person_file, mode="r", encoding="utf-8-sig")
    csvr = csv.reader(fh, delimiter="\t")
    last_int = 1
    person_ids = {}

    next(csvr)
    for persondata in csvr:
        person_ids[persondata[0]] = persondata[1]
        last_int += 1

    fh.close()
    return person_ids, last_int

def load_person_ids(person_file, person_ids, mappingrules, use_input_person_ids, person_number=1, delim=","):
    fh = open(person_file, mode="r", encoding="utf-8-sig")
    csvr = csv.reader(fh, delimiter=delim)
    person_columns = {}
    person_col_in_hdr_number = 0
    reject_count = 0

    personhdr = next(csvr)
    print(personhdr)

    # Make a dictionary of column names vs their positions
    for col in personhdr:
        person_columns[col] = person_col_in_hdr_number
        person_col_in_hdr_number += 1

    birth_datetime_source, person_id_source = mappingrules.get_person_source_field_info("person")
    print("Load Person Data {0}, {1}".format(birth_datetime_source, person_id_source))
    person_col = person_columns[person_id_source]

    for persondata in csvr:
        if not valid_value(persondata[person_columns[person_id_source]]):
            reject_count += 1
            continue
        if not valid_date_value(persondata[person_columns[birth_datetime_source]]):
            reject_count += 1
            continue
        if persondata[person_col] not in person_ids:
            if use_input_person_ids == "N":
                person_ids[persondata[person_col]] = str(person_number)
                person_number += 1
            else:
                person_ids[persondata[person_col]] = str(persondata[person_col])
    fh.close()

    return person_ids, reject_count

@click.command(help="Perform OMOP Mapping given a python configuration file.")
@click.option("--rules",
              help="input json file containing all the mapping rules to be applied")
@click.option("--pyconf","--conf",
              required=True,
              help="Run with a python configuration file")
@click.option("--object","objects",
              multiple=True,
              help="Only run specific objects.")
@click.option("--type",
              default='csv',
              type=click.Choice(['csv']),
              help="specify the type of inputs, the default is .csv inputs")
@click.option("--use-profiler",
              is_flag=True,
              help="turn on saving statistics for profiling CPU and memory usage")
@click.option("--output-folder",
              default=None,
              help="define the output folder where to dump csv files to")
@click.option("-nc","--number-of-rows-per-chunk",
              default=None,
              type=int,
              help="choose to chunk running the data into nrows")
@click.option("-np","--number-of-rows-to-process",
              default=None,
              type=int,
              help="the total number of rows to process")
@click.argument("inputs",
                #help="give a list of input files to process, and/or an input directory",
                nargs=-1)
@click.pass_context
def run_pyconfig(ctx,rules,pyconf,inputs,objects,
        output_folder,type,use_profiler,
        number_of_rows_per_chunk,
        number_of_rows_to_process):

    object_list = list(objects)
    if len(object_list) == 0:
        object_list = None

    if type != 'csv':
        raise NotImplementedError("Can only handle inputs that are .csv so far")

    #check if exists
    if any('*' in x for x in inputs):
        data_dir = os.path.dirname(carrot.__file__)
        data_dir = f'{data_dir}{os.path.sep}data{os.path.sep}'

        new_inputs = []
        for i,x in enumerate(inputs):
            if not os.path.exists(x):
                new_inputs.extend(glob.glob(f"{data_dir}{os.path.sep}{x}"))
            else:
                new_inputs.append(x)
        inputs = new_inputs

    inputs = list(inputs)
    for x in inputs:
        if os.path.isdir(x):
            inputs.remove(x)
            inputs.extend(glob.glob(f'{x}{os.path.sep}*.csv'))

    #convert the list into a map between the filename and the full path
    inputs = {
        os.path.basename(x):x
        for x in inputs
    }

    if output_folder is None:
        output_folder = f'{os.getcwd()}{os.path.sep}output_data{os.path.sep}'

    outputs = carrot.tools.create_csv_store(output_folder=output_folder)


    if not inputs:
        raise Exception('no inputs defined!')

    inputs = tools.load_csv(inputs,
                            rules=rules,
                            chunksize=number_of_rows_per_chunk,
                            nrows=number_of_rows_to_process)

    available_classes = tools.get_classes()
    if pyconf not in available_classes:
        ctx.invoke(list_classes)
        raise KeyError(f"cannot find config {pyconf}. Run 'carrot map py list' to see available classes.")

    module = __import__(available_classes[pyconf]['module'],fromlist=[pyconf])
    defined_classes = [
        m[0]
        for m in inspect.getmembers(module, inspect.isclass)
        if m[1].__module__ == module.__name__
    ]
    #should only be running one class anyway
    defined_class = defined_classes[0]
    cls = getattr(module,defined_class)

    #build a class object
    cdm = cls(inputs=inputs,
              outputs=outputs,
              use_profiler=use_profiler)
    #run it
    cdm.process(object_list)

@click.command(help="run as a Graphical User Interface (GUI)")
@click.pass_context
def gui(ctx):
    import PySimpleGUI as sg

    carrot_theme = {'BACKGROUND': 'white',
                       'TEXT': '#000000',
                       'INPUT': '#c4c6e2',
                       'TEXT_INPUT': '#000000',
                       'SCROLL': '#c4c6e2',
                       'BUTTON': ('white', '#475da7','#3DB28C'),
                       'PROGRESS': ('#01826B', '#D0D0D0'),
                       'BORDER': 1,
                       'SLIDER_DEPTH': 1,
                       'PROGRESS_DEPTH': 0}

    # Add your dictionary to the PySimpleGUI themes
    sg.theme_add_new('carrot', carrot_theme)
    sg.theme('carrot')

    _dir = os.path.dirname(os.path.abspath(carrot.__file__))
    data_dir = f"{_dir}{os.path.sep}data{os.path.sep}"

    layout = [
        [sg.Image(f'{data_dir}logo.png'),sg.T("CO-CONNECT: Dataset2CDM",font = ("Roboto", 25))],
        [sg.T('Select the rules json:')],
        [sg.Input(key='_RULES_'), sg.FilesBrowse(initial_folder=os.getcwd())],
        [sg.T('Select the input CSVs:')],
        [sg.Input(key='_INPUTS_'), sg.FilesBrowse(initial_folder=os.getcwd())],
        [sg.T('Select an output folder:')],
        [sg.Input(key='_OUTPUT_',default_text='.'), sg.FolderBrowse(initial_folder=os.getcwd())],
        #[sg.Checkbox("Mask the person_id",key="_MASK_PERSON_ID_",default=False)],
        #[[sg.T('Change the default data chunksize:'),
        #  sg.Slider(range=(0,1000000),
        #            default_value=100000,
        #            resolution=10000,
        #            orientation='horizontal')]],
        [sg.OK('Run'), sg.Cancel(button_color=('white','#3DB28C'))]
    ]

    font = ("Roboto", 15)

    window = sg.Window('carrot', layout, font=font)
    while True:
        event, values = window.Read()

        if event == 'Cancel' or event == None:
            break

        output_folder = values['_OUTPUT_']
        if output_folder == '':
            output_folder = None

        rules = values['_RULES_']
        if rules == '':
            sg.Popup(f'Error: please select a rules file')
            continue
        elif len(rules.split(';'))>1:
            sg.Popup(f'Error: only select one file for the rules!')
            continue

        inputs = values['_INPUTS_']
        if inputs == '':
            sg.Popup(f'Error: please select at least one file or directory for the inputs')
            continue
        inputs = inputs.split(';')

        #mask_person_id = values['_MASK_PERSON_ID_']

        try:
            #ctx.invoke(run,rules=rules,inputs=inputs,output_folder=output_folder,mask_person_id=mask_person_id)
            ctx.invoke(run,rules=rules,inputs=inputs,output_folder=output_folder)
            sg.Popup("Done!")
        except Exception as err:
            sg.popup_error("An exception occurred!",err)

        break

    window.close()

@click.group(help="Commands for using python configurations to run the ETL transformation.")
def py():
    pass

py.add_command(make_class,"make")
py.add_command(register_class,"register")
py.add_command(list_classes,"list")
py.add_command(remove_class,"remove")
py.add_command(run_pyconfig,"map")
run.add_command(py,"py")
run.add_command(map,"map")
run.add_command(mapstream,"mapstream")
run.add_command(list_classes, "list_classes")
run.add_command(load,"load")
run.add_command(analysis,"analysis")
run.add_command(format,"format")
run.add_command(merge,"merge")
run.add_command(transform,"transform")
run.add_command(gui,"gui")
run.add_command(test,"test")



import click
import os
import carrot

@click.group(help='Commands to find information about the package.')
def info():
    pass

@click.command(help="Get the data folder location")
def data_folder():
    _dir = os.path.dirname(os.path.abspath(carrot.__file__))
    data_dir = f"{_dir}{os.path.sep}data"
    print (data_dir)

@click.command(help="Get the root folder location of carrot tools")
def install_folder():
    _dir = os.path.dirname(os.path.abspath(carrot.__file__))
    print (_dir)

@click.command(help="Get the installed version of the package'")
def version():
    print (carrot.__version__)
    
info.add_command(install_folder,'install_folder')
info.add_command(data_folder,'data_folder')
info.add_command(version,'version')





import os
import click
import json
import yaml
import carrot
import pandas as pd
import numpy as np
import requests
import secrets
import random
import datetime
import time

class MissingToken(Exception):
    pass


@click.group(help='Commands to generate helpful files.')
def generate():
    pass

@click.group(help='Commands to generate synthetic data.')
def synthetic():
    pass

@click.command(help="generate synthetic data from a ScanReport ID from CCOM")
@click.option("-i","--report-id",help="ScanReport ID on the website",required=True,type=int)
@click.option("-n","--number-of-events",help="number of rows to generate",required=True,type=int)
@click.option("-o","--output-directory",help="folder to save the synthetic data to",required=True,type=str)
@click.option("--fill-column-with-values",help="select columns to fill values for",multiple=True,type=str)
@click.option("-t","--token",help="specify the carrot_token for accessing the CCOM website",type=str,default=None)
@click.option("--get-json",help="also download the json",is_flag=True)
@click.option("-u","--url",help="url endpoint for the CCOM website to ping",
              type=str,
              default="https://ccom.azurewebsites.net")
def ccom(report_id,number_of_events,output_directory,
         fill_column_with_values,token,get_json,
         url):

    fill_column_with_values = list(fill_column_with_values)
    
    token = os.environ.get("CARROT_MAPPER_TOKEN") or token
    if token == None:
        raise MissingToken("you must use the option --token or set the environment variable carrot_TOKEN to be able to use this functionality. I.e  export carrot_TOKEN=12345678 ")

    headers = {
        "User-Agent": "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/72.0.3626.119 Safari/537.36",
        "Content-type": "application/json",
        "charset": "utf-8",
        "Authorization": f"Token {token}"
    }
    
    if get_json:
        response = requests.get(
            f"{url}/api/json/?id={report_id}",
            headers=headers
        )
        print (json.dumps(response.json()[0],indent=6))
        fname = response.json()[0]['metadata']['dataset']
        with open(f'{fname}.json', 'w') as outfile:
            print ('saving',fname)
            json.dump(response.json()[0],outfile,indent=6)
            
    response = requests.get(
        f"{url}/api/scanreporttables/?scan_report={report_id}",
        headers=headers
    )
    if response.status_code != 200:
        print ('failed to get a response')
        print (response.json())
        exit(0)
    else:
        print (json.dumps(response.json(),indent=6))
        
    for table in response.json():
        name = table['name']
        _id = table['id']
        
        #get which is the person_id and automatically fill this with incrementing values
        #so they are not all NaN in the synthetic data (because of List Truncated...)
        person_id = table['person_id']
        if person_id == None:
            continue
        
        _url = f"{url}/api/scanreportfields/?id={person_id}&fields=name"
        person_id = requests.get(
            _url, headers=headers,
            allow_redirects=True,
        ).json()[0]['name'].lstrip('\ufeff')
        
                
        _url = f"{url}/api/scanreportvaluesfilterscanreporttable/?scan_report_table={_id}&fields=value,frequency,scan_report_field"
        response = requests.get(
            _url, headers=headers,
            allow_redirects=True,
        )
        

                
        df = pd.DataFrame.from_records(response.json()).set_index('scan_report_field')        
        _url = f"{url}/api/scanreportfields/?scan_report_table={_id}&fields=id,name"
        response = requests.get(
            _url, headers=headers,
            allow_redirects=True,
        )


        print (response.content)
        res = json.loads(response.content.decode('utf-8'))
        print (res)
        id_to_col_name = {
            field['id']:field['name'].lstrip('\ufeff')
            for field in res
        }
        
        df.index = df.index.map(id_to_col_name)

        
        df_synthetic = {}
        
        for col_name in df.index.unique():
            if col_name == '': continue
            
            _df = df.loc[[col_name]]
            _df['value'].replace('',np.nan,inplace=True)

            _df = _df.dropna()
            
            
            if len(_df) > number_of_events:
                _df = _df.sample(frac=1)[:number_of_events]

            
            frequency = _df['frequency']
            total = frequency.sum()

            
            if total > 0 :
                frequency = number_of_events*frequency / total
                frequency = frequency.astype(int)
            else:
                frequency = number_of_events
                
            values = _df['value'].repeat(frequency)\
                                 .sample(frac=1)\
                                 .reset_index(drop=True)

            values.name = col_name
            df_synthetic[col_name] = values

        df_synthetic = pd.concat(df_synthetic.values(),axis=1)
        
        
        for col_name in fill_column_with_values:
            if col_name in df_synthetic.columns:
                df_synthetic[col_name] = df_synthetic[col_name].reset_index()['index']
                
        
        if not os.path.isdir(output_directory):
            os.makedirs(output_directory)
        fname = f"{output_directory}/{name}"

        df_synthetic.index = 'pk'+df_synthetic.index.astype(str)
        df_synthetic.rename_axis(person_id,inplace=True)
        #df_synthetic.set_index(df_synthetic.columns[0],inplace=True)
        print (df_synthetic)
        df_synthetic.to_csv(fname)
        print (f"created {fname} with {number_of_events} events")
        
    
@click.command(help="generate synthetic data from a ScanReport xlsx file")
@click.argument("report")
@click.option("-n","--number-of-events",help="number of rows to generate",required=True,type=int)
@click.option("-o","--output-directory",help="folder to save the synthetic data to",required=True,type=str)
@click.option("--fill-column-with-values",help="select columns to fill values for",multiple=True,type=str)
def xlsx(report,number_of_events,output_directory,fill_column_with_values):
    dfs = pd.read_excel(report,sheet_name=None)
    sheets_to_process = list(dfs.keys())[2:-1]

    for name in sheets_to_process:
        df = dfs[name]
        columns_to_make = [
            x
            for x in df.columns
            if 'Frequency' not in x and 'Unnamed' not in x
        ]

        df_synthetic = {}
        for col_name in columns_to_make:
            i_col = df.columns.get_loc(col_name)
            df_stats = df.iloc[:,[i_col,i_col+1]].dropna()

            if not df_stats.empty:
                frequency = df_stats.iloc[:,1]
                frequency = number_of_events*frequency / frequency.sum()
                frequency = frequency.astype(int)

                values = df_stats.loc[df_stats.index.repeat(frequency)]\
                                 .iloc[:,0]\
                                 .sample(frac=1)\
                                 .reset_index(drop=True)
                df_synthetic[col_name] = values
            else:
                df_synthetic[col_name] = df_stats.iloc[:,0]
                
        df_synthetic = pd.concat(df_synthetic.values(),axis=1)

        for col_name in fill_column_with_values:
            if col_name in df_synthetic.columns:
                df_synthetic[col_name] = df_synthetic[col_name].reset_index()['index']
                
        
        if not os.path.isdir(output_directory):
            os.makedirs(output_directory)
        fname = f"{output_directory}/{name}"
        #ext = fname[-3:]
        df_synthetic.set_index(df_synthetic.columns[0],inplace=True)
        print (df_synthetic)
        df_synthetic.to_csv(fname)
        print (f"created {fname} with {number_of_events} events")


synthetic.add_command(xlsx,"from_xlsx")
synthetic.add_command(ccom,"from_ccom")

@click.command(help="generate a synthetic CDM from a yaml file")
@click.argument("config")
def synthetic_cdm(config):
    with open(config, 'r') as stream:
        config = yaml.safe_load(stream)

    dfs = {}
    order = sorted(config.items(), key=lambda x: x[0] != 'person')
    for destination_table_name,destination_table in order:
        n = destination_table['n']
        obj = carrot.cdm.get_cdm_class(destination_table_name)()
        columns = destination_table['columns']
        for column,spec in columns.items():
            if 'range' in spec:
                _range = spec['range']
                _min = 0
                if 'min' in _range:
                    _min = _range['min']
                _max = _min + n
                x = range(_min,_max)
                x = pd.Series(x)
            elif 'random' in spec:
                _map = spec['random']
                _df = pd.Series(_map).to_frame()
                x = _df.index.to_series().repeat(_df[0]*n).sample(frac=1).reset_index(drop=True)
            elif 'map' in spec:
                _map = spec['map']
                col_to_map,_map = list(_map.items())[0]
                x = obj[col_to_map].series.map(_map)
            elif 'gaus' in spec:
                _range = spec['gaus']
                mu = _range['mean']
                sigma = (_range['max'] - _range['min']).total_seconds()/5
                mu = time.mktime(mu.timetuple())
                x = [datetime.date.fromtimestamp(x) for x in np.random.normal(mu,sigma,n)]
                x = pd.Series(x)
            elif 'person' in spec:
                x = dfs['person']['person_id'].sample(n,replace=True).sort_values().reset_index(drop=True)
            else:
                print (spec)
                exit(0)

            x.name = column
            print (x)
            obj[column].series = x

        df = obj.get_df()
        print (df.dropna(axis=1))
        dfs[destination_table_name] = df
    
synthetic.add_command(synthetic_cdm,"cdm")


generate.add_command(synthetic,"synthetic")


        
@click.command(help="generate a python configuration for the given table")
@click.argument("table")
@click.argument("version")
def cdm(table,version):
    data_dir = os.path.dirname(carrot.__file__)
    data_dir = f'{data_dir}{os.path.sep}data{os.path.sep}'
    data_dir = f'{data_dir}{os.path.sep}cdm{os.path.sep}BCLINK_EXPORT{os.path.sep}'
    
    #load the details of this cdm objects from the data files taken from OHDSI GitHub
    # - set the table (e.g. person, condition_occurrence,...)  as the index
    #   so that all values associated with the object (name) can be retrieved
    # - then set the field (e.g. person_id, birth_datetime,,) to help with future lookups
    # - just keep information on if the field is required (Yes/No) and what the datatype is (INTEGER,..)
    cdm = pd.read_csv(f'{data_dir}{version}{os.path.sep}export-{table.upper()}.csv',
                      encoding="ISO-8859-1",sep='\t')\
                      .set_index('DESCRIPTION')

    for index,row in cdm.iterrows():
        required = row['REQUIRED'] == "Yes"
        dtype = row['TYPE']
        length = row['LENGTH']
        key = row['KEY']
        
        if not np.isnan(length):
            dtype = f"{dtype}{int(length)}"

        if not np.isnan(key):
            extra = ', pk=True'
        else:
            extra = ''
            
        string = f'self.{index} = DestinationField(dtype="{dtype}", required={required} {extra})'
        print (string)
        
generate.add_command(cdm,"cdm")


@click.command(help="generate a hash token to be used as a salt")
@click.option("length","--length",default=64)
def salt(length):
    salt = secrets.token_hex(length)
    click.echo(salt)

generate.add_command(salt,"salt")



def report_to_xlsx(report,f_out):
    with pd.ExcelWriter(f_out) as writer:  
        for table in report:
            table_name = table['table']
            total = []
            for field in table['fields']:
                field_name = field['field']
                values = field['values']
                data = pd.DataFrame.from_records(values)
                columns = [field_name,'Frequency']
                if data.empty:
                    data = pd.DataFrame(columns=columns)
                else:
                    data.columns=columns
                    total.append(data)
            df = pd.concat(total,axis=1)
            click.echo(df)
            df.to_excel(writer, sheet_name=table_name, index=False)



@click.command(help="generate scan report json from input data")
@click.option("max_distinct_values","--max-distinct-values",
              default=10,
              type=int,
              help='specify the maximum number of distinct values to include in the ScanReport.')
@click.option("min_cell_count","--min-cell-count",
              default=5,
              type=int,
              help='specify the minimum number of occurrences of a cell value before it can appear in the ScanReport.')
@click.option("rows_per_table","--rows-per-table",
              default=100000,
              type=int,
              help='specify the maximum of rows to scan per input data file (table).')
@click.option("randomise","--randomise",
              is_flag=True,
              help='randomise rows')
@click.option("as_type","--as-type","--save-as",
              default=None,
              type=click.Choice(['xlsx','json','latex']),
              help='save the report as a json or xlsx (whiteRabbit style).')
@click.option("f_out","--output-file-name",
              default=None,
              help='specify the output file name (to be used with --save-as)')
@click.option("name","--name",
              required=True,
              help='give a name to the report')
@click.argument("inputs",
                nargs=-1)
def report(inputs,name,max_distinct_values,min_cell_count,rows_per_table,randomise,as_type,f_out):
    skiprows = None
        
    data = []
    for fname in inputs:
        #get the name of the data table
        table_name = os.path.basename(fname)
        #load as a pandas dataframe
        #load it and preserve the original data (i.e. no NaN conversions)
        df = pd.read_csv(fname,
                         dtype=str,
                         keep_default_na=False,
                         nrows=rows_per_table,
                         skiprows=skiprows)
        if randomise:
            df = df.sample(frac=1)

        #get a list of all column (field) names
        column_names = df.columns.tolist()
        fields = []
        #loop over all colimns
        for col in column_names:
            #value count the columns
            series = df[col].value_counts()
            #reduce the size of the value counts depending on specifications of max distinct values 
            if max_distinct_values>0 and len(series)>=max_distinct_values:
                series = series.iloc[:max_distinct_values]

            #if the min cell count is set, remove value counts that are below this threshold
            if not min_cell_count is None:
                series = series[series >= min_cell_count]

            #convert into a frequency instead of value count
            series = (series/len(df)).rename('frequency').round(4) 

            #convert the values to a dictionary 
            frame = series.to_frame()
                        
            values = frame.rename_axis('value').reset_index().to_dict(orient='records')
            #record the value (frequency counts) for this field
            fields.append({'field':col,'values':values})

        meta = {
            'dataset':name,
            'nscanned':len(df),
            'max_distinct_values':max_distinct_values,
            'min_cell_count':min_cell_count
        }
        data.append({'table':table_name,'fields':fields, 'meta':meta})

    if as_type == 'json':
        f_out = f_out if f_out != None else 'ScanReport.json'
        click.echo(json.dumps(data,indent=6))
        with open(f_out, 'w') as f:
            json.dump(data, f,indent=6)
    elif as_type == 'xlsx':
        f_out = f_out if f_out != None else 'ScanReport.xlsx'
        report_to_xlsx(data,f_out)
    elif as_type == 'latex':
        for d in data:
            d = d['fields']
            for d in d:
                df = pd.DataFrame(d['values'])
                if len(df)>0:
                    print (df.to_latex(index=False))
    else:
        click.echo(json.dumps(data,indent=6))
            
generate.add_command(report,"report")


@click.command(help="generate a mapping rules json based on a json scan report")
#@click.option("max_distinct_values","--max-distinct-values",
#              default=10,
#              type=int,
#              help='specify the maximum number of distinct values to include in the ScanReport.')
@click.argument("report")
def mapping_rules(report):

    person_ids = ['ID','PersonID','person_id','Study','Study_ID','CHI','LINKNO']
    date_events = ['date','date_','date_of','time','occurrence','age']
    from difflib import SequenceMatcher
    def get_similar_score(a, b):
        return SequenceMatcher(None, a.lower(), b.lower()).ratio()
    def get_similar_scores(cols,matcher):
        return {b:max([get_similar_score(a,b) for a in matcher]) for b in cols }
    
    domain_lookup = {'Condition':'condition_occurrence','Observation':'observation','Gender':'person'}
    #date_lookup = carrot.cdm.get_cdm_class(destination_
    #print (dir(date_lookup))
    #exit(0)
    def create_hde_from_value(concept,value,table,field,person_id,date_event):
        domain_id = concept['domain_id']
        name = concept['concept_name']
        concept_id = concept['concept_id']
        destination_table = domain_lookup[concept['domain_id']]
        obj = carrot.cdm.get_cdm_class(destination_table)()
        date_events = [k for k,v in obj.get_field_dtypes().items() if v == 'Timestamp']
        date_destination = date_events[0]
        obj = {
            destination_table : {
                name : {
                    "person_id":{
                        "source_table":table,
                        "source_field":person_id,
                    },
                    date_destination:{
                        "source_table":table,
                        "source_field":date_event,
                    },
                    f"{domain_id.lower()}_source_value":{
                        "source_table":table,
                        "source_field":field,
                    },
                    f"{domain_id.lower()}_concept_id":{
                        "source_table":table,
                        "source_field":field,
                        "term_mapping": {
                            value:concept_id
                        }
                    }
                }
            }
        }
        return obj
    
    def find_person_id_and_date_event(fields):
        person_id = [k for k,v in sorted(get_similar_scores(fields,person_ids).items(),key=lambda x: x[1],reverse=True)][0]
        date_event = [k for k,v in sorted(get_similar_scores(fields,date_events).items(),key=lambda x: x[1],reverse=True)][0]
        return person_id,date_event

    def collapse(hdes):
        retval = {}
        for hde in hdes:
            for destination_table,objs in hde.items():
                if destination_table not in retval:
                    retval[destination_table] = {}
                    
                for name,obj in objs.items():
                    if name not in retval[destination_table]:
                         retval[destination_table][name] : {}
                    else:
                        raise Exception(f"{name} already exists")

                    retval[destination_table][name] = obj
        return retval
    
    report = carrot.tools.load_json(report)
    hdes = []
    for table in report:
        fields = [x['field'] for x in table['fields']]
        person_id,date_event = find_person_id_and_date_event(fields)
        for field in table['fields']:
            for value in field['values']:
                concepts = value.get('concepts')
                if not concepts:
                    continue
                for concept in concepts:
                    hde = create_hde_from_value(concept,
                                                value.get('value'),
                                                table['table'],
                                                field['field'],
                                                person_id,
                                                date_event)
                    hdes.append(hde)
    cdm = collapse(hdes)

    name = report[0]['meta']['dataset']
    rules = {
        'metadata':{
            "date_created": str(datetime.datetime.now()),
            "dataset": name
        },
        'cdm':cdm
    }
    click.echo(json.dumps(rules,indent=6))
        

generate.add_command(mapping_rules,"mapping-rules")

@click.command(help="Generate synthetic data from the json format of the scan report")
@click.option("-n","--number-of-events",help="number of rows to generate",required=True,type=int)
@click.option("-o","--output-directory",help="folder to save the synthetic data to",required=True,type=str)
@click.option("--fill-column-with-values",help="select columns to fill values for",multiple=True,type=str)
@click.argument("f_in")
def synthetic_from_json(f_in,number_of_events,output_directory,fill_column_with_values):
    fill_column_with_values = list(fill_column_with_values)
    report = json.load(open(f_in))
    for table in report:
        table_name = table['table']
        fields = table['fields']
        df_synthetic = {}
        for field in fields:
            field_name = field['field']
            values = field['values']
            df = pd.DataFrame.from_records(values)
            if len(df) == 0:
                values = pd.Series([])
            else:
                frequency = df['frequency']
                frequency = number_of_events*frequency / frequency.sum()
                frequency = frequency.astype(int)
                values = df['value'].repeat(frequency).sample(frac=1).reset_index(drop=True)
            values.name = field_name
            
            df_synthetic[field_name] = values
                
        df_synthetic = pd.concat(df_synthetic.values(),axis=1)
        for col_name in fill_column_with_values:
            if col_name in df_synthetic.columns:
                df_synthetic[col_name] = df_synthetic[col_name].reset_index()['index']

        df_synthetic.set_index(df_synthetic.columns[0],inplace=True)

        if not os.path.isdir(output_directory):
            os.makedirs(output_directory)
        fname = f"{output_directory}/{table_name}"

        click.echo(df_synthetic)
        df_synthetic.to_csv(fname)
        click.echo(f"created {fname} with {number_of_events} events")


synthetic.add_command(synthetic_from_json,"json")


            




import pandas as pd
from ...common import DestinationTable, DestinationField

class ProcedureOccurrence(DestinationTable):
    """
    CDM Procedure Occurrence object class
    """

    name = 'procedure_occurrence'
    def __init__(self,name=None):
        self.procedure_occurrence_id     = DestinationField(dtype="Integer"   , required=False , pk=True)
        self.person_id                   = DestinationField(dtype="Integer"   , required=True )
        self.procedure_concept_id        = DestinationField(dtype="Integer"   , required=True )
        self.procedure_date              = DestinationField(dtype="Date"      , required=False )
        self.procedure_datetime          = DestinationField(dtype="Timestamp" , required=True )
        self.procedure_type_concept_id   = DestinationField(dtype="Integer"   , required=False )
        self.modifier_concept_id         = DestinationField(dtype="Integer"   , required=False )
        self.quantity                    = DestinationField(dtype="Integer"   , required=False )
        self.provider_id                 = DestinationField(dtype="Integer"   , required=False )
        self.visit_occurrence_id         = DestinationField(dtype="Integer"   , required=False )
        self.procedure_source_value      = DestinationField(dtype="Text50"    , required=False )
        self.procedure_source_concept_id = DestinationField(dtype="Integer"   , required=False )
        self.qualifier_source_value      = DestinationField(dtype="Text50"    , required=False )

        if name is None:
            name = hex(id(self))
        super().__init__(name,self.name)


    def get_df(self,**kwargs):
        """
        Overload/append the creation of the dataframe, specifically for the procedure_occurrence objects
        * procedure_concept_id is required to be not null
          this can happen when spawning multiple rows from a person
          we just want to keep the ones that have actually been filled

        Returns:
           pandas.Dataframe: output dataframe
        """

        df = super().get_df(**kwargs)

        if self.automatically_fill_missing_columns == True:
            if df['procedure_date'].isnull().all():
                df['procedure_date'] = self.tools.get_date(df['procedure_datetime'])

        return df



import os
import json
from .omopcdm import OmopCDM

class MappingRules:

    def __init__(self, rulesfilepath, omopcdm):
        self.rules_data = self.load_json(rulesfilepath)
        self.omopcdm = omopcdm
        
        self.parsed_rules = {}
        self.outfile_names = {}

        self.dataset_name = self.get_dsname_from_rules()

    def load_json(self, f_in):
        """
        """
        if os.path.exists(f_in):
            data = json.load(open(f_in))
        else:
            try:
                data = json.loads(f_in)
            except Exception as err:
                raise FileNotFoundError(f"{f_in} not found. Or cannot parse as json")

        return data
    
    def dump_parsed_rules(self):
        return(json.dumps(self.parsed_rules, indent=2))

    def get_dsname_from_rules(self):
        dsname = "Unknown"

        if "metadata" in self.rules_data:
            if "dataset" in self.rules_data["metadata"]:
                dsname = self.rules_data["metadata"]["dataset"]

        return dsname

    def get_dataset_name(self):
        return self.dataset_name

    def get_all_outfile_names(self):
        file_list = []
        
        for outfilename in self.rules_data["cdm"]:
            file_list.append(outfilename)

        return file_list

    def get_all_infile_names(self):
        file_list = []

        for outfilename, conditions in self.rules_data["cdm"].items():
            for outfield, source_field in conditions.items():
                for source_field_name, source_data in source_field.items():
                    if "source_table" in source_data:
                        if source_data["source_table"] not in file_list:
                            file_list.append(source_data["source_table"])

        return file_list
        
    def get_infile_data_fields(self, infilename):
        data_fields_lists = {}

        outfilenames, outdata = self.parse_rules_src_to_tgt(infilename)

        for outfilename in outfilenames:
            data_fields_lists[outfilename] = []

        for key, outfield_data in outdata.items():
            keydata = key.split("~")
            outfile = keydata[-1]
            for outfield_elem in outfield_data:
                for infield, outfields in outfield_elem.items():
                    for outfield in outfields:
                        outfielddata = outfield.split("~")
                        if self.omopcdm.is_omop_data_field(outfile, outfielddata[0]):
                            if infield not in data_fields_lists[outfile]:
                                data_fields_lists[outfile].append(infield)

        return data_fields_lists

    def get_infile_date_person_id(self, infilename):
        outfilenames, outdata = self.parse_rules_src_to_tgt(infilename)
        datetime_source = ""
        person_id_source = ""

        for key, outfield_data in outdata.items():
            keydata = key.split("~")
            outfile = keydata[-1]
            for outfield_elem in outfield_data:
                for infield, outfield_list in outfield_elem.items():
                    #print("{0}, {1}, {2}".format(outfile, infield, str(outfield_list)))
                    for outfield in outfield_list:
                        if outfield in self.omopcdm.get_omop_datetime_fields(outfile):
                            datetime_source = infield
                        if outfield == self.omopcdm.get_omop_person_id_field(outfile):
                            person_id_source = infield

        return datetime_source, person_id_source

    def get_person_source_field_info(self, tgtfilename):
        """
        Specific discovery of input data field names for 'person' in these rules
        """
        birth_datetime_source = None
        person_id_source = None
        if tgtfilename in self.rules_data["cdm"]:
            source_rules_data = self.rules_data["cdm"][tgtfilename]
            for rule_name, rule_fields in source_rules_data.items():
                if "birth_datetime" in rule_fields:
                    birth_datetime_source = rule_fields["birth_datetime"]["source_field"]
                if "person_id" in rule_fields:
                    person_id_source = rule_fields["person_id"]["source_field"]

        return birth_datetime_source, person_id_source

    def parse_rules_src_to_tgt(self, infilename):
        """
        Parse rules to produce a map of source to target data for a given input file
        """
        if infilename in self.outfile_names and infilename in self.parsed_rules:
            return self.outfile_names[infilename], self.parsed_rules[infilename]
        outfilenames = []
        outdata = {}

        for outfilename, rules_set in self.rules_data["cdm"].items():
            for datatype, rules in rules_set.items():
                key, data = self.process_rules(infilename, outfilename, rules)
                if key != "":
                    if key not in outdata:
                        outdata[key] = []
                    outdata[key].append(data)
                    if outfilename not in outfilenames:
                        outfilenames.append(outfilename)

        self.parsed_rules[infilename] = outdata
        self.outfile_names[infilename] = outfilenames
        return outfilenames, outdata

    def process_rules(self, infilename, outfilename, rules):
        """
        Process rules for an infile, outfile combination
        """
        outkey = ""
        data = {}
        plain_key = ""
        term_value_key = ""

        for outfield, source_info in rules.items():
            if source_info["source_field"] not in data:
                data[source_info["source_field"]] = []
            if source_info["source_table"] == infilename:
                if "term_mapping" in source_info:
                    if type(source_info["term_mapping"]) is dict:
                        for inputvalue, term in source_info["term_mapping"].items():
                            term_value_key = infilename + "~" + source_info["source_field"] + "~" + str(inputvalue) + "~" + outfilename
                            data[source_info["source_field"]].append(outfield + "~" + str(source_info["term_mapping"][str(inputvalue)]))
                    else:
                        plain_key = infilename + "~" + source_info["source_field"] + "~" + outfilename
                        data[source_info["source_field"]].append(outfield + "~" + str(source_info["term_mapping"]))
                else:
                    data[source_info["source_field"]].append(outfield)
        if term_value_key != "":
            return term_value_key, data

        return plain_key, data



class Metrics():
    def __init__(self, dataset_name, log_threshold=0):
        self.datasummary={}
        self.allcounts={}
        self.log_data=""
        self.dataset_name=dataset_name
        self.log_threshold = log_threshold

    def get_new_mapstream_counts(self):
        counts = {}
        counts["input_count"] = 0
        counts["invalid_persids"] = 0
        counts["invalid_dates"] = 0
        counts["invalid_source_fields"] = 0
        counts["output_count"] = 0

        return counts

    def add_data(self, desttablename, increment):
        """
        add_data(self, destination table, data increment)
        Apply the contents of a data increment to the stored self.datasummary
        """
        name = increment["name"]
        for datakey, dataitem in increment.items():
            if datakey == "valid_person_id":
                dkey = "NA" + "." + desttablename + "." + name + "." + datakey
                self.add_counts_to_summary(dkey, dataitem)
            elif datakey == "person_id":
                dkey = "NA" + "." + desttablename + "." + name + "." + datakey
                self.add_counts_to_summary(dkey, dataitem)
            elif datakey == "required_fields":
                for fieldname in dataitem:
                    prfx = "NA"
                    if "source_files" in increment:
                        if fieldname in increment["source_files"]:
                            prfx = self.get_prefix(increment["source_files"][fieldname]["table"])
                            dkey = prfx + "." + desttablename + "." + name + "." + fieldname
                            self.add_counts_to_summary(dkey, dataitem[fieldname])

    def get_prefix(self, fname):
        return fname.split(".")[0]

    def add_counts_to_summary(self, dkey, count_block):
        if dkey not in self.datasummary:
            self.datasummary[dkey] = {}
        for counttype in count_block:
            if counttype not in self.datasummary[dkey]:
                self.datasummary[dkey][counttype] = 0
            self.datasummary[dkey][counttype] += int(count_block[counttype])

    def increment_key_count(self, dkey, count_type):
        """
        Intended to work with the mapstream functions
        """
        if dkey not in self.datasummary:
            self.datasummary[dkey] = {}
        if count_type not in self.datasummary[dkey]:
            self.datasummary[dkey][count_type] = 0
        self.datasummary[dkey][count_type] += 1

    def get_summary(self):
        summary_str = "source\ttablename\tname\tcolumn name\tbefore\tafter content check\tpct reject content check\tafter date format check\tpct reject date format\n"

        for dkey in self.datasummary:
            #print(dkey)
            source, tablename, name, colname = dkey.split('.')
            before_count = int(self.datasummary[dkey]["before"])
            after_count = int(self.datasummary[dkey]["after"])
            after_pct = (float)(before_count - after_count) * 100 / before_count
            summary_str += source + "\t" + tablename + "\t" + name + "\t" + colname + "\t" + str(before_count) + "\t" + str(after_count) + "\t" + "{0:.3f}".format(after_pct) + "\t"
            if "after_formatting" in self.datasummary[dkey]:
                after_format_count = int(self.datasummary[dkey]["after_formatting"])
                after_format_pct = (float)(after_count - after_format_count) * 100 / after_count
                summary_str += str(after_format_count) + "\t" + "{0:.3f}".format(after_format_pct) + "\n"
            else:
                summary_str += "NA\tNA\n"

        return summary_str

    def get_data_summary(self):
        return self.datasummary

    def get_mapstream_summary(self):
        summary_str = "dsname\tsource\tsource_field\ttarget\tconcept_id\tadditional\tincount\tinvalid_persid\tinvalid_date\tinvalid_source\toutcount\n"

        for dkey in sorted(self.datasummary):
            try:
                source, fieldname, tablename, concept_id, additional = dkey.split('~')
            except ValueError:
                print("get_mapstream_summary - ValueError: {0}".format(dkey))
                break
              
            source = self.get_prefix(source)
            dvalue = self.datasummary[dkey]

            input_count = "0"
            if "input_count" in dvalue:
                input_count = str(dvalue["input_count"])

            invalid_person_ids = "0"
            if "invalid_person_ids" in dvalue:
                invalid_person_ids = str(dvalue["invalid_person_ids"])

            invalid_source_fields = "0"
            if "invalid_source_fields" in dvalue:
                invalid_source_fields = str(dvalue["invalid_source_fields"])

            invalid_date_fields = "0"
            if "invalid_date_fields" in dvalue:
                invalid_date_fields = str(dvalue["invalid_date_fields"])

            output_count = "0"
            if "output_count" in dvalue:
                output_count = str(dvalue["output_count"])

            #if (int(output_count) >= self.log_threshold) or (tablename == "person"):
            if (int(output_count) >= self.log_threshold):
              summary_str += self.dataset_name + "\t" + source + "\t" + fieldname + "\t" + tablename + "\t" + concept_id + "\t" + additional +"\t" + input_count + "\t" + invalid_person_ids + "\t" + invalid_date_fields + "\t" + invalid_source_fields + "\t" + output_count + "\n"

        return summary_str

    def add_log_data(self, msg):
        self.log_data += msg + "\n"

    def get_log_data(self):
        return self.log_data
    



import os
import time
import threading
import psutil
import pandas as pd
from carrot.tools.logger import Logger
        

class Profiler(Logger):
    def __init__(self,name=None,interval=0.1):
        
        if name == None:
            name = self.__class__.__name__
        else:
            name = f"{self.__class__.__name__}_{name}"
            

        #retrieve the process id for the current run
        self.pid = os.getpid()
        #create a psutil instance to montior this
        self.py = psutil.Process(self.pid)
        #set the interval (seconds) of how often to check the cpu and memory
        self.interval = interval
        self.logger.info(f"tracking {self.pid} every {self.interval} seconds")
        #count the number of cpus the computer running this process has
        self.cpu_count = psutil.cpu_count()
        self.logger.info(f"{self.cpu_count} cpus available")
        #initiate a threaded function
        #that will run in a separate process and can monitor CPU/memory in the background
        self.th = threading.Thread(target=self.track)

        #init some global variables
        self.tracking = []
        self.init_time = time.time()
        self._stop = False
        self._df = None
        
    def start(self):
        #start the thread
        self.logger.info("starting profiling")
        self.th.start()

    def stop(self):
        #stop the thread
        self._stop = True
        self.th.join()
        self.logger.info("finished profiling")

    def get_df(self):
        #build a little dataframe for cpu/memory v.s. time,
        #if it has not been built already
        if self._df is None:
            self._df = pd.DataFrame(self.tracking)
        return self._df
        
    def summary(self):
        #print the dataframe created for cpu/memory v.s. time
        self.logger.info(self.get_df())
        
    def track(self):
        """
        Main function to profile CPU and memory usage
        """
        #while the program has been told to profile the usage
        while self._stop == False:
            #from the current process, calculate the current memory usage (in GB)
            memory = self.py.memory_info()[0]/2.**30
            #also calculate the CPU % in use at this epoch in time
            cpu = self.py.cpu_percent() / self.cpu_count
            #calcuate the current time - time since the start of the process
            current_time = time.time() - self.init_time
            #log the data
            info = {'time[s]':current_time,'memory[GB]':memory,'cpu[%]':cpu}
            self.tracking.append(info)
            #sleep the number of seconds requested
            time.sleep(self.interval)

        #once finished, call the summary function
        self.summary()



from ...common import DestinationTable, DestinationField

class Person(DestinationTable):
    """
    CDM Person object class
    """
    name = 'person'
    def __init__(self,name=None):
        self.person_id                   = DestinationField(dtype="Integer"   , required=True , pk=True)
        self.gender_concept_id           = DestinationField(dtype="Integer"   , required=True )
        self.year_of_birth               = DestinationField(dtype="Integer"   , required=False )
        self.month_of_birth              = DestinationField(dtype="Integer"   , required=False )
        self.day_of_birth                = DestinationField(dtype="Integer"   , required=False )
        self.birth_datetime              = DestinationField(dtype="Timestamp" , required=True )
        self.race_concept_id             = DestinationField(dtype="Integer"   , required=False )
        self.ethnicity_concept_id        = DestinationField(dtype="Integer"   , required=False )
        self.location_id                 = DestinationField(dtype="Integer"   , required=False )
        self.provider_id                 = DestinationField(dtype="Integer"   , required=False )
        self.care_site_id                = DestinationField(dtype="Integer"   , required=False )
        self.person_source_value         = DestinationField(dtype="Text50"    , required=False )
        self.gender_source_value         = DestinationField(dtype="Text50"    , required=False )
        self.gender_source_concept_id    = DestinationField(dtype="Integer"   , required=False )
        self.race_source_value           = DestinationField(dtype="Text50"    , required=False )
        self.race_source_concept_id      = DestinationField(dtype="Integer"   , required=False )
        self.ethnicity_source_value      = DestinationField(dtype="Text50"    , required=False )
        self.ethnicity_source_concept_id = DestinationField(dtype="Integer"   , required=False )

        if name is None:
            name = hex(id(self))
        super().__init__(name,self.name)

    def finalise(self,df,**kwargs):
        """
        Overload the finalise function here for any specifics for the person table
        """
        df = super().finalise(df,**kwargs)
        return df


    def get_df(self,**kwargs):
        """
        Overload/append the creation of the dataframe, specifically for the person objects
        * year_of_birth is automatically converted to a year (int)
        * month_of_birth is automatically converted to a month (int)
        * day_of_birth is automatically converted to a day (int)
        * birth_datetime is automatically coverted to a datatime
        

        Returns:
           pandas.Dataframe: output dataframe
        """
        df = super().get_df(**kwargs)
        if self.automatically_fill_missing_columns == True:
            if df['year_of_birth'].isnull().all():
                df['year_of_birth'] = self.tools.get_year(df['birth_datetime'])

            if df['month_of_birth'].isnull().all():
                df['month_of_birth'] = self.tools.get_month(df['birth_datetime'])

            if df['day_of_birth'].isnull().all():
                df['day_of_birth'] = self.tools.get_day(df['birth_datetime'])
                
        return df



from carrot.tools.bclink_helpers import BCLinkHelpers
from .local import LocalDataCollection
from carrot.io.common import DataBrick
import io
import pandas as pd
import time

class BCLinkDataCollection(LocalDataCollection):
    def __init__(self,bclink_settings,**kwargs):
            
        self.logger.info('setup bclink collection')
        self.bclink_helpers = BCLinkHelpers(**bclink_settings)
        
        super().__init__(**kwargs)
        self.job_ids = []

    def finalise(self):
        self.logger.info("finalising, waiting for jobs to finish")
        self.logger.info(f"job_ids to wait for: {self.job_ids}")

        #print (self.get_output_folder())

        running_jobs = self.job_ids
        while True:
            running_jobs = [j for j in running_jobs if not self.bclink_helpers.check_logs(j)]
            if len(running_jobs)==0:
                break
            self.logger.info(f"Waiting for {running_jobs} to finish")
            time.sleep(5)
         
        self.logger.info(f"done!")

    def retrieve(self):
        tables = self.bclink_helpers.get_table_map()
        for name in tables:
            df = self.bclink_helpers.get_table(name)
            b = DataBrick(df,name=name)
            self[name] = b
   

    def write(self,*args,**kwargs):
        f_out = super().write(*args,**kwargs)
        destination_table = args[0]
        self.load(f_out,destination_table)

    def load(self,f_out,destination_table):
        job_id = self.bclink_helpers.load_table(f_out,destination_table)
        if job_id:
            self.job_ids.append(job_id)
    
    def load_indexing(self):
        indexer = self.bclink_helpers.get_indicies()
        if indexer:
            self.logger.info(f'retrieved {indexer}')
        return indexer

    def load_global_ids(self):
        data = self.bclink_helpers.get_global_ids()
        if not data:
            return
        if len(data.splitlines()) == 0:
            return

        sep = self.get_separator()
        data = io.StringIO(data)
        df_ids = pd.read_csv(data,
                             sep=sep).set_index('TARGET_SUBJECT')['SOURCE_SUBJECT']
        return df_ids.to_dict()



import click

@click.group(help='Command group for configuring runs with airflow')
def airflow():
    pass

@click.command(help='Create dags for usage with airflow')
def make_dags():
    import carrot.workflows.airflow as airflow_helpers
    reports = airflow_helpers.get_reports()
    for report_id,report_name in reports.items():
        click.echo(f"creating {report_id} {report_name}")
        airflow_helpers.create_template(report_name,report_id)

@click.command(help='Create dags for usage with airflow')
def trigger_dags():
    import carrot.workflows.airflow as airflow_helpers
    airflow_helpers.trigger_etl_dags()
    

airflow.add_command(make_dags,'make_dags')
airflow.add_command(trigger_dags,'trigger_dags')






import pandas as pd
import io
import time
import json
import os
import carrot
from carrot.tools.logger import Logger
from .bash_helpers import BashHelpers
from carrot.cdm.objects import get_cdm_tables


def get_default_global_id_name():
    return 'person_ids'

def get_default_tables():
    _dict = {k:k for k in get_cdm_tables().keys()}
    _dict[get_default_global_id_name()] = get_default_global_id_name()
    return _dict


class BCLinkHelpersException(Exception):
    pass

class BCLinkHelpers(BashHelpers,Logger):

    def __init__(self,user='bclink',clean=False,check=True,gui_user='data',database='bclink',dry_run=False,tables=get_default_tables()):
        super().__init__(dry_run=dry_run)

        self.report = []
        self.user = user
        self.gui_user = gui_user
        self.database = database
        self.table_map = tables
        if self.table_map == None:
            raise BCLinkHelpersException("Table map between the dataset id and the OMOP tables must be defined")

        if check:
            self.check_tables()
        if clean:
            self.clean_tables()
        if check:
            self.print_summary()
        
    def get_table_map(self):
        return self.table_map

    def check_table_exists(self,table):
        query = f"SELECT EXISTS (SELECT 1 FROM information_schema.tables WHERE table_name = '{table}' )"
               
        cmd=[
            'bc_sqlselect',
            f'--user={self.user}',
            f'--query={query}',
            self.database
        ]
        stdout,_ = self.run_bash_cmd(cmd)
        if stdout == None:
            return  True
        return bool(int(stdout.splitlines()[1]))

    def check_tables(self):
        for table,table_name in self.table_map.items():
            exists = self.check_table_exists(table_name)
            if exists:
                self.logger.info(f"{table_name} ({table}) already exists --> all good")
            else:
                self.logger.error(f"{table_name} which is to be used for {table} does not exist!")
                #self.create_table(table_name,table)
                raise BCLinkHelpersException(f"{table_name} does not exist")

    def create_table(self,table_name,table):
        table_upper = table.upper()
        table_name_upper = table_name.upper()
        cmd = [
            'dataset_tool', 
            '--create',
            f'--table={table_name}',
            f'--setname={table_name_upper}',
            f'--user={self.gui_user}',
            f'--form={table_upper}',
            self.database
        ]
        stdout,stderr = self.run_bash_cmd(cmd)
        if stdout:
            self.logger.info(stdout)

    def create_tables(self):
        for table,table_name in self.table_map.items():
            exists = self.check_table_exists(table_name)
            if exists:
                self.logger.info(f"{table_name} ({table}) already exists, not creating")
                continue
            self.create_table(table_name,table)
          
    def get_table(self,table):
        query = f"SELECT * FROM {table}"
               
        cmd=[
            'bc_sqlselect',
            f'--user={self.user}',
            f'--query={query}',
            self.database
        ]
        stdout,_ = self.run_bash_cmd(cmd)
        if stdout == None:
            return None

        df = pd.read_csv(io.StringIO(stdout),sep='\t')
        df.columns = [x.lower() for x in df.columns]
        #df = df.drop('batch',axis=1)
        return df
       
  
    def get_bclink_table(self,table):
        if table in self.table_map:
            return self.table_map[table]
        elif table == "global_ids":
            return self.global_ids
        
        raise Exception(f"Request look up ofr table {table} which is unknown")


    def drop_duplicates(self,table):
        fields = self.get_fields(table)
        pk = fields[0]
        duplicates = self.get_duplicates(table)
        if len(duplicates) == 0:
            self.logger.info('no duplicates detected')
            return duplicates

        duplicates = ','.join([str(x) for x in duplicates])
        
        cmd=[
            'bc_sqlselect',
            f'--user={self.user}',
            f'--query=DELETE FROM {table} WHERE {pk} IN ({duplicates})',
            self.database
        ]         
        stdout,stderr = self.run_bash_cmd(cmd)
        return duplicates

    def get_duplicates(self,table):
        fields = self.get_fields(table)
        pk = fields[0]
        batch = fields[-1]
        fields = ",".join(fields[1:-1])
        
        cmd=[
            'bc_sqlselect',
            f'--user={self.user}',
            f'--query=SELECT array_agg({pk}) as duplicates FROM {table} GROUP BY {fields} HAVING COUNT(*)>1',
            self.database
        ]
        stdout,stderr = self.run_bash_cmd(cmd)
        if stdout == None:
            return [] 
        duplicates = [
            sorted([int(x) for x in dups[1:-1].split(",")])[1:]
            for dups in stdout.splitlines()[1:]
        ]
        duplicates = sorted(list(set([item for sublist in duplicates for item in sublist])))

        return duplicates

    def get_pk(self,table):
        query = f"SELECT column_name FROM INFORMATION_SCHEMA. COLUMNS WHERE table_name = '{table}' LIMIT 1 "
        cmd = [
            'bc_sqlselect',
            f'--user={self.user}',
            f'--query={query}', 
            self.database
        ]

        stdout,stdin = self.run_bash_cmd(cmd)
        if stdout == None:
            return 'person_id'
        else:
            pk = stdout.splitlines()[1]
            self.logger.info(f"got pk {pk}")
            return pk
            
    def get_fields(self,table):
        query = f"SELECT * FROM {table} LIMIT 1;"
        cmd = [
            'bc_sqlselect',
            f'--user={self.user}',
            f'--query={query}', 
            self.database
        ]

        stdout,stdin = self.run_bash_cmd(cmd)
        if stdout == None:
            return ['person_id','birth_datetime']
        else:
            return stdout.splitlines()[0].split("\t")
                      
    def get_last_index(self,table):
        pk = self.get_pk(table)
        query=f"SELECT {pk} FROM {table} ORDER BY -{pk} LIMIT 1; "
        cmd = [
            'bc_sqlselect',
            f'--user={self.user}',
            f'--query={query}', 
            self.database
        ]

       
        stdout,stderr = self.run_bash_cmd(cmd)
        if stdout == None:
            return 0
        else:
            last_index = int(stdout.splitlines()[1])
            self.logger.debug(f"Last index for {pk} in table {table} = {last_index}")
            return last_index 
                   
    
    def get_indicies(self):
        the_dict = {k:v for k,v in self.table_map.items() if not k == get_default_global_id_name()}
        reverse = {v:k for k,v in the_dict.items() }
        retval = {}
        for table in the_dict.values():
            count=['bc_sqlselect',f'--user={self.user}',f'--query=SELECT count(*) FROM {table}',self.database]
                            
            stdout,stdin = self.run_bash_cmd(count)
            if stdout == None:
                self.get_last_index(table) 
            else:
                counts = int(stdout.splitlines()[1])
                if counts > 0 :
                    retval[reverse[table]] = self.get_last_index(table) + 1

        return retval

    def check_logs(self,job_id,table=None,bclink_table=None):
        cover = f'/data/var/lib/bcos/download/data/job{job_id}/cover.{job_id}'
        if not os.path.exists(cover):
            return False
       
        cmd = f"cat {cover}"
        stdout,stderr = self.run_bash_cmd(cmd)
        if stdout == None:
            return False

        job_id=str(job_id)
        
        report = {
            'job_id':job_id,
            'table':table,
            'bclink_table':bclink_table
        }
             
        
        for msg in stdout.splitlines():
            if 'data row(s) discarded,' in msg:
                self.logger.critical(msg)
                report['dropped_rows'] = msg
            elif 'new row' in msg:
                report['new_rows'] = msg
            elif '>>> From:' in msg:
                report['From'] = msg.split('>>> From:')[1]
            elif '>>> To:' in msg:
                report['To'] = msg.split('>>> To:')[1]
            else:
                self.logger.info_v2(msg)
            
        if report not in self.report:
            self.report.append(report)
        

        return True
        
    def clean_table(self,table):
        clean = f'datasettool2 delete-all-rows {table} --database={self.database}'
        stdout,stderr = self.run_bash_cmd(clean)
        
        if stdout == None and stderr==None:
            return

        for msg in stderr.splitlines():
            self.logger.warning(msg)
                   
    def clean_tables(self,tables=None):
        for table in self.table_map.values():
            if tables is not None:
                if not table in tables:
                    continue

            self.logger.info(f"Cleaning table {table}")
            self.clean_table(table)
                   
    def get_table_jobs(self,table,head=1):
        cmd = f'datasettool2 list-updates --dataset={table} --user={self.gui_user} --database={self.database}'
        status,_ = self.run_bash_cmd(cmd)
        if status == None:
            return

        info = pd.read_csv(io.StringIO(status),
                           sep='\t',
                           usecols=['BATCH',
                                    'UPDDATE',
                                    'UPD_COMPLETION_DATE',
                                    'JOB',
                                    'STATUS',
                                    'ACTION'])
        if head is not None:
            info = info.head(head)
        return info
   
    def get_global_ids(self):
        name = get_default_global_id_name()
        if name not in self.table_map.keys():
            self.logger.warning(f"No table for getting existing person ids ({name}) has been defined")
            return None
        global_ids = self.table_map[name]
   
        query=f"SELECT * FROM {global_ids} "
        cmd=['bc_sqlselect',f'--user={self.user}',f'--query={query}',self.database]
        
        stdout,stderr = self.run_bash_cmd(cmd)
        #if stdout == None:
        #    return None
        #if len(stdout.splitlines()) == 0:
        #    return None

        return stdout#io.StringIO(stdout)
        
        #df_ids = pd.read_csv(io.StringIO(stdout),
        #                     sep='\t').set_index("SOURCE_SUBJECT")
       # 
       # df_ids.to_csv(f_out,sep='\t')
       # return f_out

    def check_global_ids(self,output_directory,chunksize=10):

        if not self.global_ids:
            return True

        data_file = f'{output_directory}/person.tsv'
        if not os.path.exists(data_file):
            self.logger.warning(f"{output_directory}/person.tsv file does not exist")
            self.logger.warning("skipping global id check")
            return True

        data = carrot.tools.load_csv({"ids":f"{output_directory}/global_ids.tsv"},
                                        sep='\t',
                                        chunksize=100)
        
        while True:
            query=f"select exists(select 1 from {self.global_ids} where TARGET_SUBJECT in ({_list}) )"
            cmd=['bc_sqlselect',f'--user={self.user}',f'--query={query}',self.database]
            stdout,stderr = self.run_bash_cmd(cmd)
            if stdout == None:
                exists = False   
            else:    
                exists = bool(int(stdout.splitlines()[1]))
            
            if exists:
                #if any target IDs (hashed!) already exist in the table... 
                #check if the pairing is different
                _list = ','.join([f"('{s}','{t}')" for s,t in data["ids"].values])
                self.logger.debug("getting IDs that overlap")
                query=f"select SOURCE_SUBJECT,SOURCE_SUBJECT from {self.global_ids} where (SOURCE_SUBJECT,TARGET_SUBJECT) in ({_list}) "
                cmd=['bc_sqlselect',f'--user={self.user}',f'--query={query}',self.database]
                stdout,stderr = self.run_bash_cmd(cmd)
                info = pd.read_csv(io.StringIO(stdout),
                                   sep='\t').set_index("SOURCE_SUBJECT")
                self.logger.error(info)
                return False

            try:
                data.next()
            except StopIteration:
                break
   
        return True
        
   
    def load_global_ids(self,output_directory):

        if self.global_ids == None:
            return

        data_file = f'{output_directory}{os.path.sep}global_ids.tsv'
        if not os.path.exists(data_file):
            #raise FileExistsError(
            self.logger.error(f"Cannot find global_ids.tsv in output directory: {output_directory}")
            return 
           
        cmd = ['dataset_tool', '--load',f'--table={self.global_ids}',f'--user={self.gui_user}',
               f'--data_file={data_file}','--support','--bcqueue',self.database]
        
        stdout,stderr = self.run_bash_cmd(cmd)
        if not stdout == None:
            for msg in stdout.splitlines():
                self.logger.info(f"submitted job to bclink queue: {msg}")

        table_name = self.global_ids
        stats = self.get_table_jobs(table_name)
        if stats is None:
            #is a dry run, just test this
            self.check_logs(0)
        else:
            job_id = stats.iloc[0]['JOB']
            while True:
                stats = self.get_table_jobs(table_name)
                self.logger.info(stats)
                self.logger.info(f"Getting log for {table_name} id={job_id}")
                success = self.check_logs(job_id,'global_ids',table_name)
                if success:
                    break
                else:
                    self.logger.debug(f"Didn't find the log for {table_name} id={job_id} yet, job still running. Trying again in 5 seconds..")
                    time.sleep(5)

    def load_table(self,f_out,destination_table):

        try:
            tablename = self.table_map[destination_table]
        except KeyError:
            self.logger.error(f"table {destination_table} unknown in {self.table_map.keys()}")
            return
        
        if not os.path.exists(f_out):
            self.logger.error(f"Cannot find {f_out} to load to bclink.")
            return
                
        cmd = ['dataset_tool', '--load',f'--table={tablename}',f'--user={self.gui_user}',
               f'--data_file={f_out}','--support','--bcqueue',self.database]
        
        stdout,stderr = self.run_bash_cmd(cmd)
        if not stdout == None:
            for msg in stdout.splitlines():
                self.logger.info(f"submitted job to bclink queue: {msg}")
            
        stats = self.get_table_jobs(tablename)
        if stats is None:
            #is a dry run, just test this
            #self.check_logs(0)
            pass
        else:
            job_id = stats.iloc[0]['JOB']
            self.logger.info(f"running job {job_id}")
            #self.check_logs(job_id)
            return job_id

    def load_tables(self,output_directory,tables_to_process=None):
        for table,tablename in self.table_map.items():
            if tables_to_process is not None:
                if table not in tables_to_process:
                    continue

            data_file = f'{output_directory}/{table}.tsv'
            if not os.path.exists(data_file):
                #raise FileExistsError(
                self.logger.error(f"Cannot find {table}.tsv in output directory: {output_directory}")
                continue

            cmd = ['dataset_tool', '--load',f'--table={tablename}',f'--user={self.gui_user}',
                   f'--data_file={data_file}','--support','--bcqueue',self.database]
            
            stdout,stderr = self.run_bash_cmd(cmd)
            if not stdout == None:
                for msg in stdout.splitlines():
                    self.logger.info(f"submitted job to bclink queue: {msg}")


        for table,table_name in self.table_map.items():
            if tables_to_process is not None:
                if table not in tables_to_process:
                    continue
            self.logger.debug(f"Checking jobs submitted for {table_name}")
            stats = self.get_table_jobs(table_name)
            if stats is None:
                #is a dry run, just test this
                self.check_logs(0)
            else:
                job_id = stats.iloc[0]['JOB']
                while True:
                    stats = self.get_table_jobs(table_name)
                    self.logger.info(stats)
                    self.logger.info(f"Getting log for {table_name} id={job_id}")
                    success = self.check_logs(job_id,table,table_name)
                    if success:
                        break
                    else:
                        self.logger.warning(f"Didn't find the log for {table_name} id={job_id} yet, job still running.")
                        time.sleep(1)

        self.print_summary()

    def print_report(self):
        if self.report:
            self.logger.info(json.dumps(self.report,indent=6))

    def print_summary(self):
        info = {}
        for table,table_name in self.table_map.items():
            if table_name == None:
                continue
            cmd=['bc_sqlselect',f'--user={self.user}',f'--query=SELECT count(*) FROM {table_name}',self.database]
            
            stdout,stderr = self.run_bash_cmd(cmd)
            if stdout == None:
                continue
                
            count = stdout.splitlines()[1]
            info[table] = {'bclink_table':table_name,
                           'nrows':count}            
        
        if info:
            self.logger.info("======== BCLINK SUMMARY ========")
            self.logger.info(json.dumps(info,indent=6))
              

    def remove_table(self,fname):
        self.logger.info(f"Called remove_table on {fname}")
        table = os.path.splitext(os.path.basename(fname))[0].split('.')[0]
        data = carrot.tools.load_csv({table:{'fields':[0],'file':fname}},
                                        sep='\t',
                                        chunksize=1000)


        if not table in self.table_map:
            raise KeyError(f"{table} doesnt exist in bclink mapping")

        bc_table = self.table_map[table]

                    
        pk = self.get_pk(bc_table)
        self.logger.debug(f"will remove {bc_table} using primary-key={pk}")
            
        while True:
            indices_to_delete = ','.join(data[table].iloc[:,0].values)
            self.logger.debug(f"removing {len(indices_to_delete)} indices from {bc_table}")
            query=f"DELETE FROM {bc_table} WHERE {pk} IN ({indices_to_delete}) "
            cmd=['bc_sqlselect',f'--user={self.user}',f'--query={query}',self.database]
            
            stdout,stderr = self.run_bash_cmd(cmd)
            
            try:
                data.next()
            except StopIteration:
                break



from ...common import DestinationTable, DestinationField

class Death(DestinationTable):
    """
    CDM Death object class
    """
    name = 'death'
    def __init__(self,name=None):
        self.person_id               = DestinationField(dtype="Integer"   , required=False , pk=True)
        self.death_date              = DestinationField(dtype="Date"      , required=False )
        self.death_datetime          = DestinationField(dtype="Timestamp" , required=False )
        self.death_type_concept_id   = DestinationField(dtype="Integer"   , required=False )
        self.cause_concept_id        = DestinationField(dtype="Integer"   , required=False )
        self.cause_source_value      = DestinationField(dtype="Text50"    , required=False )
        self.cause_source_concept_id = DestinationField(dtype="Integer"   , required=False )

        if name is None:
            name = hex(id(self))
        super().__init__(name,self.name)

    def finalise(self,df,**kwargs):
        """
        Overload the finalise function here for any specifics for the person table
        """
        df = super().finalise(df,**kwargs)
        return df


    def get_df(self,**kwargs):
        """
        Returns:
           pandas.Dataframe: output dataframe
        """
        df = super().get_df(**kwargs)
        if self.automatically_fill_missing_columns == True:
            pass
                
        return df



import pandas as pd
from ...common import DestinationTable, DestinationField

class Measurement(DestinationTable):
    """
    CDM Measurement object class
    """
    
    name = 'measurement'
    def __init__(self,name=None):
        self.measurement_id                = DestinationField(dtype="Integer"   , required=True  , pk=True)
        self.person_id                     = DestinationField(dtype="Integer"   , required=True )
        self.measurement_concept_id        = DestinationField(dtype="Integer"   , required=True )
        self.measurement_date              = DestinationField(dtype="Date"      , required=False )
        self.measurement_datetime          = DestinationField(dtype="Timestamp" , required=True )
        self.measurement_type_concept_id   = DestinationField(dtype="Integer"   , required=False )
        self.operator_concept_id           = DestinationField(dtype="Integer"   , required=False )
        self.value_as_number               = DestinationField(dtype="Float"     , required=False )
        self.value_as_concept_id           = DestinationField(dtype="Integer"   , required=False )
        self.unit_concept_id               = DestinationField(dtype="Integer"   , required=False )
        self.range_low                     = DestinationField(dtype="Float"     , required=False )
        self.range_high                    = DestinationField(dtype="Float"     , required=False )
        self.provider_id                   = DestinationField(dtype="Integer"   , required=False )
        self.visit_occurrence_id           = DestinationField(dtype="Integer"   , required=False )
        self.measurement_source_value      = DestinationField(dtype="Text50"    , required=False )
        self.measurement_source_concept_id = DestinationField(dtype="Integer"   , required=False )
        self.unit_source_value             = DestinationField(dtype="Text50"    , required=False )
        self.value_source_value            = DestinationField(dtype="Text50"    , required=False )

        if name is None:
            name = hex(id(self))
        super().__init__(name,self.name)

    def get_df(self,**kwargs):
        """
        Overload/append the creation of the dataframe, specifically for the measurement objects
        * measurement_concept_id is required to be not null
          this can happen when spawning multiple rows from a person
          we just want to keep the ones that have actually been filled
        
        Returns:
           pandas.Dataframe: output dataframe
        """

        df = super().get_df(**kwargs)
        if self.automatically_fill_missing_columns == True:
            if df['measurement_date'].isnull().all():
                df['measurement_date'] = self.tools.get_date(df['measurement_datetime'])
            if df['value_as_number'].isnull().all():
                df['value_as_number'] = pd.to_numeric(df['measurement_source_value'],errors='coerce').astype('Float64')
        
        return df
